/*
    libview64 image C64 decoding library
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: libview64.c 92 2016-07-17 07:26:04Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include "libview64.h"
#include <string.h>

static view64_screen *new_screen(unsigned short int w, unsigned short int h, unsigned int n)
{
    size_t len;
    view64_screen *screen;
    len = w * h + 1; /* overflow for shifted pictures */

    screen = (view64_screen *)malloc(sizeof(view64_screen));
    if (screen == NULL) return NULL;
    screen->w = w;
    screen->h = h;
    screen->b = NULL;
    screen->n = 0;
    screen->flags = 0;
    do {
        screen->d[screen->n] = (unsigned char *)malloc(len);
        if (screen->d[screen->n] == NULL) {
            return NULL;
        }
        screen->n++;
    } while (screen->n < n);
    return screen;
}

void view64_free_screen(view64_screen *screen)
{
    while (screen->n) free(screen->d[--screen->n]);
    free(screen->b);
    free(screen);
}

static void hires_pic2(unsigned char *d, const unsigned char *bitmap, const unsigned char *color)
{
    int x, y, x2, y2;
    unsigned char c[2], b, *p;
    const int screen_w = 320;

    p = d;
    for (y = 0; y < 25; y++) {
        for (x = 0; x < 40; x++)
        {
            c[0] = *(color++);
            c[1] = c[0] / 16;
            c[0] &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                b = bitmap[y2];
                for (x2 = 0; x2 < 8; x2++)
                {
                    p[x2] = c[b >> 7];
                    b <<= 1;
                }
                p += screen_w;
            }
            p -= screen_w * 8 - 8;
            bitmap += 8;
        }
        p += screen_w * 7;
    }
}

static view64_screen *hires_pic(const unsigned char *bitmap, const unsigned char *color, unsigned char border)
{
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = border & 15;

    hires_pic2(screen->d[0], bitmap, color);
    return screen;
}

static view64_screen *PIC_godot(const unsigned char *mem, size_t size)
{
    size_t i;
    int x, y, x2, y2, db, p;
    unsigned char c, *d;
    const static unsigned char palette[16] = {
        0, 6, 9, 11, 2, 4, 8, 12, 14,
        10, 5, 15, 3, 7, 13, 1
    };
    int w, h;
    view64_screen *screen;

    if (mem[0] == 0) { 
        if (size < 2) return NULL;
        w = 320; h = 200;
        i = 2; p = -1; 
    } else {
        if (size < 4) return NULL;
        if (memcmp(mem, "GOD", 3)) return NULL;
        p = (mem[3] & 0x20) ? 0xad : -1;
        switch (mem[3] | 0x20) {
        case '0':
            w = 320; h = 200;
            i = 4;
            break;
        case '1':
            if (size < 8) return NULL;
            w = mem[6] * 8; h = mem[7] * 8;
            i = 8;
            break;
        default: 
            return NULL;
        }
    }
    screen = new_screen(w, h, 1);
    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = 0;

    db = 0;
    for (y = 0; y < h; y += 8) {
        for (x = 0; x < w; x += 8) {
            d = screen->d[0] + y * w + x;
            for (y2 = 0; y2 < 8; y2++) {
                for (x2 = 0; x2 < 8; x2 += 2) {
                    if (db > 0) {
                        db--;
                    } else {
                        c = i < size ? mem[i++] : 0;
                        if (c == p) {
                            db = i < size ? mem[i++] : 0;
                            if (!db) db = 256;
                            c = i < size ? mem[i++] : 0;
                            db--;
                        }
                    }
                    d[x2] = palette[c >> 4];
                    d[x2 + 1] = palette[c & 15];
                }
                d += w;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_vbm(const unsigned char *mem, size_t size)
{
    int x, y, x2, w, h;
    unsigned char c, *d;
    size_t p;
    view64_screen *screen;
    int compressed;
    int invert;

    if (size < 8) return NULL;
    w = mem[4] * 256 + mem[5];
    h = mem[6] * 256 + mem[7];
    switch (mem[3]) {
    case 2:
        p = 8;
        compressed = 0;
        invert = 0x00;
        break;
    case 3:
        if (size < 18) return NULL;
        p = 18 + mem[16] * 256 + mem[17];
        compressed = mem[8];
        invert = 0xff;
        break;
    default:
        return NULL;
    }

    screen = new_screen(w, h, 1);
    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = 1;
    d = screen->d[0];

    if (compressed) {
        int repeat = 0;
        for (y = 0; y < h; y++) {
            for (x = 0; x < w; x += 8) {
                if (repeat > 0) {
                    repeat--;
                } else {
                    c = p < size ? mem[p++] : 0;
                    if (c == mem[9]) {
                        c = p < size ? mem[p++] : 0;
                        repeat = p < size ? mem[p++] : 0;
                        repeat--;
                    } else if (c == mem[10] || c == mem[11]) {
                        c = (c == mem[10]) ? 0x00 : 0xff;
                        repeat = p < size ? mem[p++] : 0;
                        repeat--;
                    } else if (c == mem[12] || c == mem[13]) {
                        c = (c == mem[12]) ? 0x00 : 0xff;
                        repeat = 1;
                    } 
                }
                for (x2 = 0; x2 < (x == (w & ~7) ? (w & 7) : 8); x2++) {
                    *d++ = (c & (128 >> x2)) ? 1 : 0;
                }
            }
        }
    } else {
        for (y = 0; y < h; y++) {
            for (x = 0; x < w; x += 8) {
                c = p < size ? (mem[p++] ^ invert) : invert;
                for (x2 = 0; x2 < (x == (w & ~7) ? (w & 7) : 8); x2++) {
                    *d++ = (c & 128) ? 0 : 1;
                    c <<= 1;
                }
            }
        }
    }
    return screen;
}

static view64_screen *PIC_charpad(const unsigned char *mem, size_t size)
{
    int tile_w, tile_h, map_h, map_w, chars, tiles;
    const unsigned char *char_data, *char_attrib, *tile_data, *tile_colours, *tile_attrib, *map_data, *m;
    unsigned char c[4], c3;
    view64_screen *screen;
    int t, x, y, w, h;
    unsigned char *d;
    int multi, expanded, tile_system, word_tile, word_char, char_bytes;

    if (size < 20) return NULL;
    if (memcmp(mem, "CTM", 3)) return NULL;
    c[0] = mem[4] & 15;
    c[1] = mem[5] & 15;
    c[2] = mem[6] & 15;
    c3 = mem[7] & 15;
    switch (mem[3]) {
    case 1:
        multi = mem[9] & 1;
        map_w = mem[10] + mem[11]*256;
        map_h = mem[12] + mem[13]*256;
        chars = mem[14] + 1;
        tiles = mem[15] + 1;
        tile_w = 4;
        tile_h = 4;
        word_tile = 0; word_char = 0;
        expanded = 0;
        tile_system = 1;
        m = mem + 20;
        char_bytes = 8;
        char_data = m; m += chars * 8;
        tile_data = m;
        if (size < m - mem + tiles * tile_h * tile_w) {
            return NULL;
        }
        for (t = 0; t < tiles; t++) {
            for (y = 0; y < tile_h; y++) {
                for (x = 0; x < tile_w; x++) {
                    if (*m++ >= chars) return NULL;
                }
            }
        }
        char_attrib = NULL;
        if (mem[8] == 2) {
            tile_attrib = m; m += tiles * tile_h * tile_w;
        } else {
            tile_attrib = NULL;
        }
        if (mem[8] == 1) {
            tile_colours = m; m += tiles;
        } else {
            tile_colours = NULL;
        }
        map_data = m;
        if (size < m - mem + map_h * map_w) {
            return NULL;
        }
        for (y = 0; y < map_h; y++) {
            for (x = 0; x < map_w; x++) {
                if (*m++ >= tiles) return NULL;
            }
        }
        break;
    case 2:
        multi = mem[9] & 1;
        map_w = mem[10] + mem[11]*256;
        map_h = mem[12] + mem[13]*256;
        chars = mem[14] + mem[15]*256 + 1;
        tiles = mem[16] + 1;
        tile_w = mem[17];
        tile_h = mem[17];
        if (!tile_w || tile_w > 5) return NULL;
        word_tile = 0; word_char = 1;
        expanded = 0;
        tile_system = 1;
        m = mem + 20;
        char_bytes = 9;
        char_data = m; m += chars * 9;
        tile_data = m;
        if (size < m - mem + tiles * tile_h * tile_w * 2) {
            return NULL;
        }
        for (t = 0; t < tiles; t++) {
            for (y = 0; y < tile_h; y++) {
                for (x = 0; x < tile_w; x++) {
                    if (m[0] + m[1]*256 >= chars) return NULL;
                    m += 2;
                }
            }
        }
        char_attrib = NULL;
        if (mem[8] == 1) {
            tile_colours = m; m += tiles;
        } else {
            tile_colours = NULL;
        }
        if (mem[8] == 2) {
            tile_attrib = m; m += tiles * tile_h * tile_w;
        } else {
            tile_attrib = NULL;
        }
        map_data = m;
        if (size < m - mem + map_h * map_w) {
            return NULL;
        }
        for (y = 0; y < map_h; y++) {
            for (x = 0; x < map_w; x++) {
                if (*m++ >= tiles) return NULL;
            }
        }
        break;
    case 3:
        multi = mem[9] & 1;
        map_w = mem[10] + mem[11]*256;
        map_h = mem[12] + mem[13]*256;
        chars = mem[14] + mem[15]*256 + 1;
        tiles = mem[16] + 1;
        tile_w = mem[17];
        tile_h = mem[17];
        if (!tile_w || tile_w > 5) return NULL;
        word_tile = 0; word_char = 1;
        expanded = mem[18] & 1;
        tile_system = 1;
        m = mem + 20;
        char_bytes = 8;
        char_data = m; m += chars * 8;
        if (expanded) {
            if (chars != tiles * tile_w * tile_h) return NULL;
            tile_data = NULL;
        } else {
            tile_data = m;
            if (size < m - mem + tiles * tile_h * tile_w * 2) {
                return NULL;
            }
            for (t = 0; t < tiles; t++) {
                for (y = 0; y < tile_h; y++) {
                    for (x = 0; x < tile_w; x++) {
                        if (m[0] + m[1]*256 >= chars) return NULL;
                        m += 2;
                    }
                }
            }
        }
        char_attrib = NULL;
        if (mem[8] == 2) {
            tile_attrib = m; m += tiles * tile_h * tile_w;
        } else {
            tile_attrib = NULL;
        }
        if (mem[8] == 1) {
            tile_colours = m; m += tiles;
        } else {
            tile_colours = NULL;
        }
        map_data = m;
        if (size < m - mem + map_h * map_w) {
            return NULL;
        }
        for (y = 0; y < map_h; y++) {
            for (x = 0; x < map_w; x++) {
                if (*m++ >= tiles) return NULL;
            }
        }
        break;
    case 4:
        tile_system = 1;
        multi = mem[9] & 1;
        chars = mem[10] + mem[11]*256 + 1;
        tiles = mem[12] + 1;
        tile_w = mem[13];
        tile_h = mem[14];
        if (!tile_w || tile_w > 8 || !tile_h || tile_h > 8) return NULL;
        word_tile = 0; word_char = 1;
        map_w = mem[15] + mem[16]*256;
        map_h = mem[17] + mem[18]*256;
        expanded = mem[19] & 1;
        m = mem + 24;
        char_bytes = 8;
        char_data = m; m += chars * 8;
        char_attrib = NULL; m += chars;
        if (expanded) {
            if (chars != tiles * tile_w * tile_h) return NULL;
            tile_data = NULL;
        } else {
            tile_data = m;
            if (size < m - mem + tiles * tile_h * tile_w * 2) {
                return NULL;
            }
            for (t = 0; t < tiles; t++) {
                for (y = 0; y < tile_h; y++) {
                    for (x = 0; x < tile_w; x++) {
                        if (m[0] + m[1]*256 >= chars) return NULL;
                        m += 2;
                    }
                }
            }
        }
        tile_attrib = (mem[8] == 2) ? m : NULL; m += tiles * tile_h * tile_w;
        if (mem[8] == 1) { // per tile colour
            tile_colours = m; m += tiles;
        } else {
            tile_colours = NULL;
        }
        map_data = m;
        if (size < m - mem + map_h * map_w) {
            return NULL;
        }
        for (y = 0; y < map_h; y++) {
            for (x = 0; x < map_w; x++) {
                if (*m++ >= tiles) return NULL;
            }
        }
        break;
    case 5:
        multi = mem[9] & 4;
        tile_system = mem[9] & 1;
        chars = mem[10] + mem[11]*256 + 1;
        if (tile_system) { // tile system
            expanded = mem[9] & 2;
            tiles = mem[12] + mem[13]*256 + 1;
            tile_w = mem[14];
            tile_h = mem[15];
            if (!tile_w || tile_w > 8 || !tile_h || tile_h > 8) return NULL;
            if (expanded && chars != tiles * tile_w * tile_h) return NULL;
        } else {
            if (mem[8] == 2) return NULL;
            expanded = 0;
            tiles = 0;
            tile_w = 1;
            tile_h = 1;
        }
        word_tile = 1; word_char = 1;
        map_w = mem[16] + mem[17]*256;
        map_h = mem[18] + mem[19]*256;
        m = mem + 20;
        char_bytes = 8;
        char_data = m; m += chars * 8;
        char_attrib = (mem[8] == 2) ? m : NULL; m += chars;
        if (expanded || !tile_system) { // expanded or no tile system
            tile_data = NULL;
        } else {
            tile_data = m;
            if (size < m - mem + tiles * tile_h * tile_w * 2) {
                return NULL;
            }
            for (t = 0; t < tiles; t++) {
                for (y = 0; y < tile_h; y++) {
                    for (x = 0; x < tile_w; x++) {
                        if (m[0] + m[1]*256 >= chars) return NULL;
                        m += 2;
                    }
                }
            }
        }
        tile_attrib = NULL;
        if (mem[8] == 1 && tile_system) { // per tile colour and tilesystem
            tile_colours = m; m += tiles;
        } else {
            tile_colours = NULL;
        }
        map_data = m;
        if (size < m - mem + map_h * map_w * 2) {
            return NULL;
        }
        if (tile_system) { // tile system
            for (y = 0; y < map_h; y++) {
                for (x = 0; x < map_w; x++) {
                    if (m[0] + m[1]*256 >= tiles) return NULL;
                    m += 2;
                }
            }
        } else {
            for (y = 0; y < map_h; y++) {
                for (x = 0; x < map_w; x++) {
                    if (m[0] + m[1]*256 >= chars) return NULL;
                    m += 2;
                }
            }
        }
        break;
    default:
        return NULL;
    }

    w = map_w * tile_w * 8;
    h = map_h * tile_h * 8;
    screen = new_screen(w, h, 1);
    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = c[0];
    d = screen->d[0];

    m = map_data;
    for (y = 0; y < map_h; y++) {
        for (x = 0; x < map_w; x++) {
            int tilenum = *m++;
            const unsigned char *tile, *attr;
            unsigned char *d2 = d + x * tile_w * 8 + (y * tile_h * 8) * w;
            int tx, ty;
            if (word_tile) tilenum += (*m++)*256;
            tile = tile_data ? &tile_data[tilenum * tile_w * tile_h * (1 + word_char)] : NULL;
            attr = tile_attrib ? &tile_attrib[tilenum * tile_w * tile_h] : NULL;
            if (tile_colours) c3 = tile_colours[tilenum] & 15;
            if (expanded) tilenum *= tile_w * tile_h;
            for (ty = 0; ty < tile_h; ty++) {
                for (tx = 0; tx < tile_w; tx++) {
                    int charnum;
                    const unsigned char *chr;
                    int cx, cy;
                    if (tile) {
                        charnum = *tile++;
                        if (word_char) charnum += (*tile++)*256;
                    } else charnum = tilenum;
                    if (attr) c3 = (*attr++) & 15;
                    chr = &char_data[charnum * char_bytes];
                    if (expanded) tilenum++;
                    if (char_attrib) c3 = char_attrib[charnum] & 15;
                    if ((c3 & 8) && multi) { // multi
                        c[3] = c3 & 7;
                        for (cy = 0; cy < 8; cy++) {
                            unsigned char ch = chr[cy];
                            for (cx = 0; cx < 8; cx += 2) {
                                d2[cx] = d2[cx + 1] = c[ch >> 6];
                                ch <<= 2;
                            }
                            d2 += w;
                        }
                    } else {
                        for (cy = 0; cy < 8; cy++) {
                            unsigned char ch = chr[cy];
                            for (cx = 0; cx < 8; cx++) {
                                d2[cx] = (ch & 128) ? c3 : c[0];
                                ch <<= 1;
                            }
                            d2 += w;
                        }
                    }
                    d2 -= w * 8 - 8;
                }
                d2 += w * 8 - tile_w * 8;
            }
        }
    }

    return screen;
}

static void paint_sprite(const unsigned char *m, unsigned char *d3, unsigned char *c, int w) {
    int x, y;
    c[2] = m[63] & 15;
    for (y = 0; y < 21; y++) {
        unsigned char *d2 = d3 + y * w;
        for (x = 0; x < 3; x++) {
            int cx;
            unsigned char ch = m[y * 3 + x];
            if (m[63] & 128) { // multi
                for (cx = 0; cx < 8; cx += 2) {
                    if (ch & 192) d2[cx] = d2[cx + 1] = c[ch >> 6];
                    ch <<= 2;
                }
            } else {
                for (cx = 0; cx < 8; cx++) {
                    if (ch & 128) d2[cx] = c[2];
                    ch <<= 1;
                }
            }
            d2 += 8;
        }
        d2 -= w * 8 - 8 * 3;
    }
}

static view64_screen *PIC_spritepad(const unsigned char *mem, size_t size)
{
    const unsigned char *m;
    unsigned char c[4];
    view64_screen *screen;
    int n, w, h, t;
    unsigned char *d;

    if (!memcmp(mem, "SPD", 3)) {
        if (mem[3] != 1) return NULL;
        n = mem[4] + 1;
        c[0] = mem[6];
        c[1] = mem[7];
        c[3] = mem[8];
        m = mem + 9;
        if (size < 9 + n * 64) return NULL;
    } else {
        if ((size & 63) != 3 || size > 256*64+3 || size < 67) return NULL;
        c[0] = mem[0];
        c[1] = mem[1];
        c[3] = mem[2];
        m = mem + 3;
        n = size / 64;
    }

    w = (n < 8 ? n : 8) * 24;
    h = ((n + 7) / 8) * 21;
    screen = new_screen(w, h, 1);
    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = c[0];
    d = screen->d[0];
    memset(d, c[0], w * h);

    for (t = 0; t < n; t++) {
        unsigned char *d2 = d + (t % 8) * 24 + (t / 8) * 21 * w;
        paint_sprite(m, d2, c, w);
        if ((m[63] & 0x10) && t + 1 < n) {
            paint_sprite(m + 64, d2, c, w);
        }
        m += 64;
    }
    return screen;
}

static void multi_pic2(unsigned char *d, const unsigned char *bitmap, const unsigned char *color1, const unsigned char *color2, unsigned char bgcolor)
{
    int x, y, x2, y2;
    unsigned char c[4], b, *p;
    const int screen_w = 320;

    c[0] = bgcolor & 15;
    p = d;
    for (y = 0; y < 25; y++) {
        for (x = 0; x < 40; x++)
        {
            c[2] = *(color1++);
            c[1] = c[2] / 16;
            c[2] &= 15;
            c[3] = *(color2++) & 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                b = bitmap[y2];
                for (x2 = 0; x2 < 8; x2 += 2)
                {
                    unsigned char ki = c[b >> 6];
                    p[x2] = ki;
                    p[x2 + 1] = ki;
                    b <<= 2;
                }
                p += screen_w;
            }
            p -= screen_w * 8 - 8;
            bitmap += 8;
        }
        p += screen_w * 7;
    }
}

static view64_screen *multi_pic(const unsigned char *bitmap, const unsigned char *color1, const unsigned char *color2, unsigned char brcolor, unsigned char bgcolor)
{
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = brcolor & 15;

    multi_pic2(screen->d[0], bitmap, color1, color2, bgcolor);
    return screen;
}

static view64_screen *mci_pic(const unsigned char *bitmap1, const unsigned char *bitmap2, const unsigned char *color1, const unsigned char *color2, const unsigned char *color3, unsigned char bgcolor, unsigned char brcolor, int noshift)
{
    int y;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 2);

    if (screen == NULL) return NULL;
    screen->br = brcolor & 15;

    multi_pic2(screen->d[0], bitmap1, color1, color3, bgcolor);
    multi_pic2(screen->d[1] + !noshift, bitmap2, color2, color3, bgcolor);
    if (!noshift) {
        screen->flags |= VIEW64_SCREEN_MCI;
        for (y = 0; y < screen_h; y++) screen->d[1][y * screen_w] = bgcolor & 15;
    }

    return screen;
}

static void fli_pic2(unsigned char *d, const unsigned char *bitmap, const unsigned char *colors, const unsigned char *color2, const unsigned char *backg, const unsigned char *profi)
{
    int x, y, x2, y2, yy;
    unsigned char c[4], b, b2, *p, cm1, cm2, cs;
    const int screen_w = 320;

    cm1 = profi[0x448];
    cm2 = profi[0x449];

    for (y = 0; y < 25; y++) {
        for (x = 0; x < 40; x++)
        {
            c[3] = color2[y * 40 + x] & 15;
            p = d + (y * screen_w + x) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                yy = y * 8 + y2;
                c[0] = backg[yy] & 15;
                if (x < 3) {
                    c[2] = 0xff;
                    cs = profi[0x280 + yy] & 0x0f;
                    c[3] = profi[0x380 + yy] >> 4;
                    b2 = profi[x + ((yy / 2) % 21) * 3 + (yy / 42) * 64 + ((yy + 1) & 2) * 0xa0];
                } else {
                    c[2] = colors[y * 40 + x + y2 * 1024];
                    cs = b2 = 0;
                }

                c[1] = c[2] / 16;
                c[2] &= 15;
                b = bitmap[(y * 40 + x) * 8 + y2];
                for (x2 = 0; x2 < 8; x2 += 2)
                {
                    unsigned char ki = c[b >> 6];
                    switch (b2 & 192) {
                    case 64: ki = cs; break;
                    case 128: ki = cm1; break;
                    case 192: ki = cm2;
                    }
                    p[x2] = ki;
                    p[x2 + 1] = ki;
                    b *= 4;
                    b2 *= 4;
                }
                p += screen_w;
            }
        }
    }
}


static view64_screen *ifli_pic(const unsigned char *bitmap1, const unsigned char *bitmap2, const unsigned char *color1, const unsigned char *color2, const unsigned char *color3, const unsigned char *bgcolors)
{
    int y;
    unsigned char profi[0x480];
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 2);

    if (screen == NULL) return NULL;
    screen->br = 0;
    screen->flags |= VIEW64_SCREEN_MCI;

    memset(profi, 0, 0x480);
    memset(profi + 0x380, 0x90, 200);
    fli_pic2(screen->d[0], bitmap1, color1, color3, bgcolors, profi);
    fli_pic2(screen->d[1] + 1, bitmap2, color2, color3, bgcolors, profi);
    for (y = 0; y < screen_h; y++) screen->d[1][y * screen_w] = bgcolors[y] & 15;
    return screen;
}

static void afli_pic(unsigned char *d, const unsigned char *bitmap, const unsigned char *colors)
{
    int x, y, x2, y2;
    unsigned char c[2], b, *p;
    const int screen_w = 320;

    p = d;
    for (y = 0; y < 25; y++) {
        for (x = 0; x < 40; x++)
        {
            for (y2 = 0; y2 < 8; y2++)
            {
                if (x >= 3) { 
                    c[0] = colors[y2 * 1024];
                    c[1] = c[0] / 16;
                    c[0] &= 15;
                    b = bitmap[y2];
                    for (x2 = 0; x2 < 8; x2++)
                    {
                        p[x2] = c[b >> 7];
                        b <<= 1;
                    }
                } else memset(p, 0, 8);
                p += screen_w;
            }
            p -= screen_w * 8 - 8;
            colors++;
            bitmap += 8;
        }
        p += screen_w * 7;
    }
}

static view64_screen *iafli_pic(const unsigned char *bitmap1, const unsigned char *colors1, const unsigned char *bitmap2, const unsigned char *colors2)
{
    int interlace = (bitmap1 != bitmap2 || colors1 != colors2);
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, interlace ? 2 : 1);

    if (screen == NULL) return NULL;
    screen->br = 0;

    afli_pic(screen->d[0], bitmap1, colors1);
    if (interlace) {
        afli_pic(screen->d[1], bitmap2, colors2);
    }
    return screen;
}

static view64_screen *fli_pic(const unsigned char *bitmap, const unsigned char *colors, const unsigned char *color2, const unsigned char *backg, const unsigned char *profi, unsigned char border)
{
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = border & 15;

    fli_pic2(screen->d[0], bitmap, colors, color2, backg, profi);
    return screen;
}

static view64_screen *PIC_fli(const unsigned char *mem) //.fli
{
    unsigned char backg[200], profi[0x480];

    memset(profi, 0, 0x480);
    memset(backg, 0, 200);

    return fli_pic(mem + 0x2402, mem + 0x0402, mem + 0x0002, backg, profi, 0);
}

static view64_screen *PIC_emc_editor(const unsigned char *mem) //.emc
{
    unsigned char backg[200], profi[0x480], bitmap[8000];
    int i;

    memset(profi, 0, 0x480);
    memset(backg, mem[0x4403] & 15, 200);
    memcpy(bitmap, mem + 0x2002, sizeof(bitmap));
    for (i = 0; i < 40; i++) {
        memset(bitmap + 0x0000 + i * 8, 0, 4);
        memset(bitmap + 0x1e04 + i * 8, 0, 4);
    }

    return fli_pic(bitmap, mem + 0x0002, mem + 0x4002, backg, profi, backg[0]);
}

static view64_screen *PIC_fli_editor(const unsigned char *mem) //.fed
{
    unsigned char profi[0x480];

    memset(profi, 0, 0x480);
    memset(profi + 0x380, 0x90, 200);

    return fli_pic(mem + 0x2502, mem + 0x0502, mem + 0x0102, mem + 0x0008, profi, 0);
}

static view64_screen *PIC_fligraph(const unsigned char *mem, size_t size) //.flg
{
    unsigned char data2[17472];
    unsigned char profi[0x480];
    int i, j;
    unsigned char key, byte, db;

    i = size;

    if (i != 17474 && i >= 18) {
        j = 0x443f;
        key = mem[2];

        for (;; ) {
            byte = mem[--i]; if (i < 18) {
                break;
            }
            if (byte == key) {
                db = mem[--i] - 1; if (i < 18 || db > j) {
                    break;
                }
                byte = mem[--i]; if (i < 18) {
                    break;
                }
                memset(&data2[j - db], byte, db + 1);
                j -= db + 1;
            } else {
                data2[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }
    } else {
        memcpy(data2, mem + 2, sizeof(data2));
    }

    memset(profi, 0, 0x480);
    memcpy(profi + 0x380, data2, 200);

    return fli_pic(data2 + 0x2500, data2 + 0x500, data2 + 0x100, data2, profi, 0);
}

static view64_screen *PIC_flip(const unsigned char *mem, size_t size) //.fbi
{
    unsigned char data2[0x4440], backg[200], profi[0x480];
    unsigned char key, byte, db;
    int i, j = -1;

    i = size;

    if (i != 17218 && i >= 18) {
        j = 0x443f;
        key = mem[2];

        for (;; ) {
            byte = mem[--i]; if (i < 18) {
                break;
            }
            if (byte == key) {
                db = mem[--i] - 1; if (i < 18 || db > j) {
                    break;
                }
                byte = mem[--i]; if (i < 18) {
                    break;
                }
                memset(&data2[j - db], byte, db + 1);
                j -= db + 1;
            } else {
                data2[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }
    }
    if (j < 0) {
        return PIC_fligraph(mem, size);
    }

    memset(profi, 0, 0x480);   // 14 kiveve utolso 7, mert ott 10.
    memset(backg, 0, 200);

    return fli_pic(data2 + 0x2500, data2 + 0x0500, data2 + 0x100, backg, profi, 0);
}

static view64_screen *PIC_fliprofi(const unsigned char *mem) //.fpr
{
    unsigned char backg[200];

    memset(backg, 0, 200);

    return fli_pic(mem + 0x2882, mem + 0x882, mem + 0x0482, backg, mem + 0x0002, 0);
}

/* start: $3aff, len: 26115, name: FFLI
 * $3aff - ident $66 
 * $3b00 - $3bc7 background colors
 * $3c00 - $3fe7 color2
 * $4000 - $5fe7 color1s
 * $6000 - $7f3f bitmap
 * $8000 - $9fe7 color1s
 * $a000 - $a0c7 background colors
 */
static view64_screen *PIC_flash_fli(const unsigned char *mem) //.ffl - no editor
{
    unsigned char profi[0x480];
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 2);

    if (screen == NULL) return NULL;
    screen->br = 0;

    memset(profi, 0, 0x480);
    memset(profi + 0x380, 0x90, 200);
    fli_pic2(screen->d[0], mem + 0x2503, mem + 0x0503, mem + 0x0103, mem + 0x0003, profi);
    fli_pic2(screen->d[1], mem + 0x2503, mem + 0x4503, mem + 0x0103, mem + 0x6503, profi);
    return screen;
}

static view64_screen *PIC_hcb(const unsigned char *mem) //.hcb
{
    const unsigned char *bitmap, *colors, *color2, *backg;
    int x, y, x2, y2, yy, i;
    unsigned char c1 = 0, c2 = 0, c3 = 0, b, *p;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = 0;

    color2 = mem + 0x0002;
    colors = mem + 0x0802;
    bitmap = mem + 0x1002;
    backg = mem + 0x2f42;
    i = 0;
    for (y = 0; y < 25; y++) {
        for (x = 0; x < 40; x++)
        {
            p = screen->d[0] + (y * screen_w + x) * 8;
            yy = y * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                if (x < 3 && (y2 & 4)) {
                    c3 = 0x00;
                    c2 = 0x0f;
                    c1 = 0x0f;
                } else if (!(y2 & 3)) {
                    c3 = color2[i + (y2 & 4) * 256] & 15;
                    c2 = colors[i + (y2 & 4) * 256];
                    c1 = c2 / 16;
                    c2 &= 15;
                }

                b = bitmap[i * 8 + y2];
                for (x2 = 0; x2 < 8; x2 += 2)
                {
                    unsigned char ki;
                    switch (b & 192) {
                    case 0: ki = backg[yy / 4] & 15; break;
                    case 64: ki = c1; break;
                    case 128: ki = c2; break;
                    default: ki = c3;
                    }
                    p[x2] = ki;
                    p[x2 + 1] = ki;
                    b *= 4;
                }
                p += screen_w;
                yy++;
            }
            i++;
        }
    }
    return screen;
}

enum rle_e {
    RLE_BYTE_LEN, RLE_LEN_BYTE, RLE_BYTE_LEN1
};

static void rle_depack(const unsigned char *mem, size_t size, unsigned char key, unsigned char *data, size_t max, enum rle_e type)
{
    unsigned char byte;
    size_t pos = 0, db;
    size_t p = 0;

    while (pos < max)
    {
        if (p >= size) break;
        byte = mem[p++]; 
        if (byte != key) {
            data[pos++] = byte;
            continue;
        }
        if (p >= size) break;
        db = byte = mem[p++];
        if (p >= size) break;
        if (type == RLE_LEN_BYTE) {
            db = mem[p++];
        } else {
            byte = mem[p++];
        }
        if (type == RLE_BYTE_LEN1) db++;
        if (!db) db = 256;
        if (pos + db > max) db = max - pos;
        memset(data + pos, byte, db);
        pos += db;
    }
}

static view64_screen *PIC_super_hires_fli(const unsigned char *mem, size_t size)
{
    unsigned char data[15872];
    unsigned char data2[0x2000 + 256];
    int x, y, x2, y2, i;
    unsigned char c0, c1, cs1, cs2, b1, b2, b3, *p;
    const static unsigned char shfsprites[64] = {
        0x80, 0x84, 0x85, 0x89, 0x8A, 0x8E, 0x8F, 0x93,
        0x94, 0x98, 0x99, 0x9D, 0x9E, 0xA2, 0xA3, 0xA7,
        0xA8, 0xAC, 0xAD, 0xB1, 0xB2, 0xB6, 0xB7, 0xBB,
        0xBC, 0xC0, 0xC1, 0xC5, 0xC6, 0xCA, 0xCB, 0xCF,
        0xD0, 0xD4, 0xD5, 0xD9, 0xDA, 0xDE, 0xDF, 0xE3,
        0xE4, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE,
        0xEF, 0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6,
        0xF7, 0x1E, 0x2E, 0x3E, 0x4E, 0x5E, 0x6E, 0x7E
    };
    const static unsigned char crest[5] = {
        'C' + 64, 'R' + 64, 'E' + 64, 'S' + 64, 'T' + 64
    };
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 3) return NULL;
    screen = new_screen(screen_w, screen_h, 1);
    if (screen == NULL) return NULL;
    screen->br = 15;

    memset(data, 255, sizeof(data));
    rle_depack(mem + 3, size - 3, mem[2], data2, 0x2000, RLE_BYTE_LEN);

    if (!memcmp(data2 + 0x1feb, crest, 5)) {
        data[0x3e8] = data2[0x1fe8];
        data[0x3e9] = data2[0x1fe9];
        i = 0;
        for (y = 0; y < 167; y++) {
            for (x = 0; x < 12; x++)
            {
                data[shfsprites[(y & 0x7) * 8 + (x / 3)] * 64 + (x % 3) + (y % 21) * 3] = data2[i];
                data[shfsprites[(y & 0x7) * 8 + (x / 3) + 4] * 64 + (x % 3) + (y % 21) * 3] = data2[0x800 + i];
                y++;
                data[0x2070 + (y & 0x7) + x * 8 + (y / 8) * 320] = data2[0x1000 + i];
                data[14 + (y & 0x7) * 1024 + x + (y / 8) * 40] = data2[0x1800 + i];
                y--; i++;
            }
        }
    } else {
        memcpy(data, mem + 2, sizeof(data));
    }

    memset(screen->d[0], 15, screen_w * screen_h);

    cs1 = data[0x3e8] & 0xf;
    cs2 = data[0x3e9] & 0xf;

    for (y = 0; y < 21; y++) {
        for (x = 14; x < 40; x++)
        {
            p = screen->d[0] + (y * screen_w + x) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                c0 = data[y * 40 + x + y2 * 1024];
                c1 = c0 / 16;
                c0 &= 15;
                b1 = data[8192 + (y * 40 + x) * 8 + y2];
                if ((x < 26) && (y2 + y * 8 > 0)) {
                    int yy = ((x - 14) % 3) + ((y * 8 + y2 - 1) % 21) * 3;
                    int yy2 = 8 * ((y2 - 1) & 7) + ((x - 14) / 3);
                    b2 = data[yy + shfsprites[yy2] * 64];
                    b3 = data[yy + shfsprites[yy2 + 4] * 64];
                } else {
                    b2 = b3 = 0;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    *p = (b2 & 128) ? cs1 : ((b3 & 128) ? cs2 : ((b1 & 128) ? c1 : c0));
                    p++;
                    b1 *= 2;
                    b2 *= 2;
                    b3 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_xfli(const unsigned char *mem, size_t size)
{
    unsigned char data[16192];
    unsigned char data2[16192];
    int x, y, x2, i, j;
    unsigned char c0, c1, cs123[16], cs, b1, b2, *p, key, db, byte;
    const static unsigned char xflisprites[64] = {
        0xE4, 0xE5, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE,
        0xEF, 0xF0, 0xF1, 0xF2, 0xF3, 0xF6, 0xF7, 0xF8,
        0x0E, 0x1E, 0x2E, 0x3E, 0x4E, 0x5E, 0x6E, 0x7E,
        0x80, 0x81, 0x85, 0x86, 0x8A, 0x8B, 0x8F, 0x90,
        0x94, 0x95, 0x99, 0x9A, 0x9E, 0x9F, 0xA3, 0xA4,
        0xA8, 0xA9, 0xAD, 0xAE, 0xB2, 0xB3, 0xB7, 0xB8,
        0xBC, 0xBD, 0xC1, 0xC2, 0xC6, 0xC7, 0xCB, 0xCC,
        0xD0, 0xD1, 0xD5, 0xD6, 0xDA, 0xDB, 0xDF, 0xE0
    };
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 2) return NULL;
    size -= 2;

    screen = new_screen(screen_w, screen_h, 1);
    if (screen == NULL) return NULL;
    screen->br = 15;

    i = size;
    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));

    if (i != sizeof(data) && i) {
        j = 0x3f3f;
        key = data[0];

        for (;; ) {
            byte = data[--i]; if (i == 0) {
                break;
            }
            if (byte == key) {
                db = data[--i] - 1; if (i == 0 || db > j) {
                    break;
                }
                byte = data[--i]; if (i == 0) {
                    break;
                }
                memset(&data2[j - db], byte, db + 1);
                j -= db + 1;
            } else {
                data2[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }

        memcpy(data, data2, sizeof(data));
    }

    memset(screen->d[0], 15, screen_w * screen_h);

    memcpy(cs123 + 1, data + 0x1fee, 10);

    p = screen->d[0] + 128;
    for (y = 0; y < 168; y++) {
        if (y >= 4 && !(y & 1)) {
            int off = (y - 4) / 2;
            cs123[data[0x3dc + (off % 28) + (off / 28) * 1024] & 15] = data[0x3c0 + (off % 28) + (off / 28) * 1024] & 0x0f;
        }
        for (x = 16; x < 40; x++)
        {
            c0 = data[(y / 8) * 40 + x + ((y + 6) & 7) * 1024];
            c1 = c0 / 16;
            c0 &= 15;
            b1 = data[8192 + ((y / 8) * 40 + x) * 8 + (y & 7)];
            if (y) {
                int yy = ((x - 16) % 3) + (((y - 1) % 42) / 2) * 3;
                int yy2 = 8 * ((y - 1) & 7) + ((x - 16) / 3);
                b2 = data[yy + xflisprites[yy2] * 64];
            } else {
                b2 = 0;
            }
            for (x2 = 0; x2 < 8; x2 += 2)
            {
                switch (b2 & 192) {
                case 64: cs = cs123[1]; break;
                case 128: cs = cs123[(x - 7) / 3]; break;
                case 192: cs = cs123[2]; break;
                default: cs = c0;
                }
                *p++ = (b1 & 128) ? c1 : cs;
                *p++ = (b1 & 64) ? c1 : cs;
                b1 *= 4; b2 *= 4;
            }
        }
        p += screen_w - 192;
    }
    return screen;
}

static void ufli_pic(unsigned char *d, const unsigned char *bitmap, const unsigned char *color, const unsigned char *sprite, const unsigned char *scolor, unsigned char bg)
{
    int x, y, x2, y2, yy;
    unsigned char c0, c1, b1, b2, *p;
    const int screen_w = 320, screen_h = 200;

    memset(d, bg, screen_w * screen_h);
    for (y = 0; y < 25; y++) {
        for (x = 3; x < 39; x++)
        {
            p = d + (y * screen_w + x) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                c0 = color[y * 40 + x + (y2 / 2) * 1024];
                c1 = c0 / 16;
                c0 &= 15;
                yy = y * 8 + y2;
                b1 = bitmap[(y * 40 + x) * 8 + y2];
                b2 = sprite[(yy / 40) * 12 * 64 + (y2 & 2) * 3 * 64 + ((yy + 1) / 2 % 21) * 3 + ((x - 3) / 2 % 3) + ((x - 3) / 6) * 64];
                if (~x & 1) {
                    b2 *= 16;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    p[x2] = (b1 & 128) ? c1 : ((b2 & 128) ? (scolor[((x - 3) / 6)] & 0xf) : c0);
                    b1 *= 2;
                    if (x2 & 1) {
                        b2 *= 2;
                    }
                }
                p += screen_w;
            }
        }
    }
}

static view64_screen *PIC_ufli(const unsigned char *mem, size_t size)
{
    unsigned char data[16384 + 256];
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 3) return NULL;
    screen = new_screen(screen_w, screen_h, 1);
    if (screen == NULL) return NULL;

    memset(data, 255, sizeof(data));
    rle_depack(mem + 3, size - 3, mem[2], data, 0x4000, RLE_BYTE_LEN);

    if (data[0x3f38] == 'C' + 64 &&
        data[0x3f39] == 'R' + 64 &&
        data[0x3f3a] == 'E' + 64 &&
        data[0x3f3b] == 'S' + 64 &&
        data[0x3f3c] == 'T' + 64 &&
        data[0x3f3d] == '\'' + 128 &&
        data[0x3f3e] == '9' + 128 &&
        data[0x3f3f] == '5' + 128) {
    } else {
        size -= 2;
        memcpy(data, mem + 2, (size < 16384) ? size : 16384);
    }
    screen->br = data[0xff1] & 0xf;

    if (data[0xfef] != 1) {
        memset(data + 0xff2, data[0xff0], 6);
    }

    ufli_pic(screen->d[0], data + 0x2000, data + 0x1000, data, data + 0xff2, data[0xff1] & 0xf);
    return screen;
}

static void mufli_pic(unsigned char *d, const unsigned char *data)
{
    int x, y, x2, y2, yy;
    unsigned char c0, c1, cs, b1, b2, *p, mc, cm1, cm2;
    const static unsigned short soff[6] = {
        0x300, 0x380, 0x700, 0x780, 0xb00, 0xb80
    };
    const int screen_w = 320, screen_h = 200;

    memset(d, data[0x1ef2] & 0xf, screen_w * screen_h);
    cm1 = data[0x1ef0] & 0xf;
    cm2 = data[0x1ef1] & 0xf;
    for (y = 0; y < 25; y++) {
        for (x = 3; x < 40; x++)
        {
            int sn = (x - 3) / 6;
            p = d + (y * screen_w + x) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                int so;
                yy = y * 8 + y2;
                if (y < 16) {
                    so = 0x1f00 + (((15 - y2 - y * 8) >> 1) & 7) * 1024;
                    c0 = data[so + y * 40 + x];
                    b1 = data[0x3f00 + (y * 40 + x) * 8 + y2];
                    if (x < 39) {
                        b2 = data[data[so + 0x3f9 + sn] * 64 + 0x1f00 + ((((yy + 47) >> 1) * 3 + ((x - 3) / 2 % 3) + (yy > 122 ? 1 : 0)) & 63)];
                    } else {
                        b2 = 0;
                    }
                } else {
                    if (y < 21) {
                        so = -0x100 + (y2 >> 1) * 1024;
                    } else {
                        so = -0x100 + (((y2 + 2) >> 1) & 3) * 1024;
                    }
                    c0 = data[so + y * 40 + x];
                    b1 = data[-0x100 + (y * 40 + x) * 8 + y2];
                    if (x < 39) {
                        x2 = data[so + 0x3f9 + sn] * 64 - 0x2100 + ((((yy + 47) >> 1) * 3 + ((x - 3) / 2 % 3) + ((yy < 165) ? 1 : 2)) & 63);
                        if (x2 >= 0) {
                            b2 = data[x2];
                        } else {
                            b2 = 0;
                        }
                    } else {
                        b2 = 0;
                    }
                }
                mc = data[0x1200 + ((yy + 1) >> 1)];
                if (x < 39) {
                    cs = data[soff[sn] + ((yy + 1) >> 1)] & 0xf;
                } else {
                    cs = 0;
                }
                c1 = c0 / 16;
                c0 &= 15;
                if (~x & 1) {
                    b2 *= 16;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    if ((mc >> sn) & 2) {
                        switch (b2 & 192) {
                        case 0: *p = (b1 & 128) ? c1 : c0; break;
                        case 64: *p = (b1 & 128) ? c1 : cm2; break;
                        case 128: *p = (b1 & 128) ? c1 : cs; break;
                        case 192: *p = (b1 & 128) ? c1 : cm1; break;
                        }
                        if ((x2 & 3) == 3) {
                            b2 *= 4;
                        }
                    } else {
                        *p = (b1 & 128) ? c1 : ((b2 & 128) ? cs : c0);
                        if (x2 & 1) {
                            b2 *= 2;
                        }
                    }
                    p++;
                    b1 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
} 

static view64_screen *PIC_mufli(const unsigned char *mem, size_t size)
{
    unsigned char data[21824];
    unsigned char data2[21824];
    int i, j;
    unsigned char key, byte, db;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 2) return NULL;
    screen = new_screen(screen_w, screen_h, 1);
    if (screen == NULL) return NULL;

    size -= 2;
    i = size;
    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));

    if (i != sizeof(data) && i) {
        j = 0x54ff;
        key = data[0];

        for (;; ) {
            byte = data[--i]; if (i == 0) {
                break;
            }
            if (byte == key) {
                db = data[--i] - 1; if (i == 0 || db > j) {
                    break;
                }
                byte = data[--i]; if (i == 0) {
                    break;
                }
                memset(&data2[j - db], byte, db + 1);
                j -= db + 1;
            } else {
                data2[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }

        memcpy(data, data2, sizeof(data));
    }
    screen->br = data[0x1ef2] & 0xf;

    mufli_pic(screen->d[0], data);
    return screen;
}

static view64_screen *PIC_nufli(const unsigned char *mem, size_t size)
{
    unsigned char data[23040];
    unsigned char data2[23040];
    int x, y, x2, y2, yy, i, j;
    unsigned char c0, c1, cs, b1, b2 = 0, b3 = 0, *p, cm1, cm2, cm3, ch, br, key, byte, db, sprcol[6];
    const static unsigned short soff[6] = {
        0x400, 0x480, 0x800, 0x880, 0xc00, 0xc80
    };
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 2) return NULL;
    screen = new_screen(screen_w, screen_h, 1);
    if (screen == NULL) return NULL;
    screen->b = (unsigned char *)malloc(screen_h * 2);
    if (screen->b == NULL) {
        free(screen);
        return NULL;
    }

    size -= 2;
    i = size;
    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));

    if (i != sizeof(data) && i > 1 && data[0] == 0xfd) {
        j = 0x59ff;
        key = data[1];

        for (;; ) {
            byte = data[--i]; if (i == 1) {
                break;
            }
            if (byte == key) {
                db = data[--i] - 1; if (i == 1 || db > j) {
                    break;
                }
                byte = data[--i]; if (i == 1) {
                    break;
                }
                memset(&data2[j - db], byte, db + 1);
                j -= db + 1;
            } else {
                data2[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }

        memcpy(data, data2, sizeof(data));
    }

    for (i = 0; i < 6; i++) {
        sprcol[i] = data[soff[i]] & 0xf;
    }
    br = data[0x1fe0];
    cm1 = data[0x1ff1] & 0xf;
    cm2 = data[0x1ff0] & 0xf;
    cm3 = data[0x1ff6] & 0xf;
    ch = data[0x1ff7] & 0xf;
    for (y = 0; y < 25; y++) {
        for (y2 = 0; y2 < 8; y2++) {
            yy = y * 8 + y2;
            p = screen->d[0] + yy * screen_w;
            screen->b[yy * 2] = br;
            if (yy) {
                for (i = 0; i < 6; i++) {
                    unsigned char cs;
                    cs = data[soff[i] + ((yy + 1) >> 1)];
                    switch (cs & 0xf0) {
                    case 0x00: //simple color
                        if (y2 & 1) {
                            sprcol[i] = cs;
                        }
                        break;
                    case 0x20: //border color
                        if (y2 & 1) {
                            br = cs & 0xf;
                        }
                        break;
                    case 0x50:
                        if ((!i && (y2 & 1)) || (i && !(y2 & 1))) {
                            cm1 = cs & 0xf;
                        }
                        break;
                    case 0x60:
                        if ((!i && (y2 & 1)) || (i && !(y2 & 1))) {
                            cm2 = cs & 0xf;
                        }
                        break;
                    case 0x70:
                        if ((!i && (y2 & 1)) || (i && !(y2 & 1))) {
                            ch = cs & 0xf;
                        }
                        break;
                    case 0xe0:
                        if ((!i && (y2 & 1)) || (i && !(y2 & 1))) {
                            cm3 = cs & 0xf;
                        }
                        break;
                    }
                }
            }
            screen->b[yy * 2 + 1] = br;
            for (x = 0; x < 40; x++)
            {
                int sn = (x - 3) / 6;
                int so;

                if (y < 16) {
                    so = 0x2000 + (((15 - y2 - y * 8) >> 1) & 7) * 1024;
                    c0 = (x > 2 || y2 < 2) ? data[so + y * 40 + x] : 0xff;
                    b1 = data[0x4000 + (y * 40 + x) * 8 + y2];
                    if (x < 3) {
                        b2 = data[data[so + 0x3f8] * 64 + 0x2000 + ((((yy + 47) >> 1) * 3 + x + (yy > 122 ? 1 : 0)) & 63)];
                        b3 = data[data[so + 0x3ff] * 64 + 0x2000 + ((((yy + 47) >> 1) * 3 + x + (yy > 122 ? 1 : 0)) & 63)];
                    } else if (x < 39) {
                        b2 = data[data[so + 0x3f9 + sn] * 64 + 0x2000 + ((((yy + 47) >> 1) * 3 + ((x - 3) / 2 % 3) + (yy > 122 ? 1 : 0)) & 63)];
                    } else {
                        b2 = 0;
                    }
                } else {
                    if (y < 21) {
                        so = (y2 >> 1) * 1024;
                    } else {
                        so = (((y2 + 2) >> 1) & 3) * 1024;
                    }
                    c0 = (x > 2 || y2 < 2) ? data[so + y * 40 + x] : 0xff;
                    b1 = data[(y * 40 + x) * 8 + y2];
                    if (x < 3) {
                        x2 = data[so + 0x3f8] * 64 - 0x2000 + ((((yy + 47) >> 1) * 3 + x + ((yy < 165) ? 1 : 2)) & 63);
                        if (x2 >= 0) {
                            b2 = data[x2];
                        }
                        x2 = data[so + 0x3ff] * 64 - 0x2000 + ((((yy + 47) >> 1) * 3 + x + ((yy < 165) ? 1 : 2)) & 63);
                        if (x2 >= 0) {
                            b3 = data[x2];
                        }
                    } else if (x < 39) {
                        x2 = data[so + 0x3f9 + sn] * 64 - 0x2000 + ((((yy + 47) >> 1) * 3 + ((x - 3) / 2 % 3) + ((yy < 165) ? 1 : 2)) & 63);
                        if (x2 >= 0) {
                            b2 = data[x2];
                        } else {
                            b2 = 0;
                        }
                    } else {
                        b2 = 0;
                    }
                }
                c1 = c0 / 16;
                c0 &= 15;
                if (x < 3) {
                    for (x2 = 0; x2 < 8; x2 += 2)
                    {
                        switch (b3 & 192) {
                        case 64: cs = cm1; break;
                        case 128: cs = cm3; break;
                        case 192: cs = cm2; break;
                        default: cs = c0; break;
                        }
                        *p++ = (b1 & 128) ? c1 : ((b2 & 128) ? ch : cs);
                        *p++ = (b1 & 64) ? c1 : ((b2 & 64) ? ch : cs);
                        b1 *= 4; b2 *= 4; b3 *= 4;
                    }
                } else if (x < 39) {
                    cs = sprcol[sn];
                    if (~x & 1) {
                        b2 *= 16;
                    }
                    for (x2 = 0; x2 < 8; x2 += 2)
                    {
                        *p++ = (b1 & 128) ? c1 : ((b2 & 128) ? cs : c0);
                        *p++ = (b1 & 64) ? c1 : ((b2 & 128) ? cs : c0);
                        b1 *= 4;
                        b2 *= 2;
                    }
                } else {
                    for (x2 = 0; x2 < 8; x2++) {
                        *p++ = (b1 & 128) ? c1 : c0;
                        b1 *= 2;
                    }
                }
            }
        }
    }
    screen->br = screen->b[0] = screen->b[1] = screen->b[2] = br;
    return screen;
}

static view64_screen *PIC_muifli(const unsigned char *mem, size_t size)
{
    unsigned char data[0xd600];
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 2) return NULL;
    screen = new_screen(screen_w, screen_h, 2);
    if (screen == NULL) return NULL;

    size -= 2;
    memcpy(data, mem + 2, (size < 44034) ? size : 44034);
    memcpy(data + 0xcd00 - 0x2100, data + 0x7700 - 0x2100, 0x2a00);

    screen->br = data[0x1ef2] & 0xf;

    mufli_pic(screen->d[0], data);
    mufli_pic(screen->d[1], data + 0x8000);
    return screen;
}

static view64_screen *PIC_uifli(const unsigned char *mem, size_t size)
{
    unsigned char data[32768 + 256], data2[32768 + 256];
    int i, j;
    unsigned char db, key, byte;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 2) return NULL;
    screen = new_screen(screen_w, screen_h, 2);
    if (screen == NULL) return NULL;

    memset(data, 255, sizeof(data));
    size -= 2;
    i = size;
    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));

    if (data[0x3f38] == 'C' + 64 &&
        data[0x3f39] == 'R' + 64 &&
        data[0x3f3a] == 'E' + 64 &&
        data[0x3f3b] == 'S' + 64 &&
        data[0x3f3c] == 'T' + 64 &&
        data[0x3f3d] == '\'' + 128 &&
        data[0x3f3e] == '9' + 128 &&
        data[0x3f3f] == '8' + 128) {
        memcpy(data + 0x4000, data, 0x4000);
    } else {
        if (i) {
            j = 0x7f3f;
            key = data[0];

            for (;; ) {
                byte = data[--i]; if (i == 0) {
                    break;
                }
                if (byte == key) {
                    db = data[--i] - 1; if (i == 0 || db > j) {
                        break;
                    }
                    byte = data[--i]; if (i == 0) {
                        break;
                    }
                    memset(&data2[j - db], byte, db + 1);
                    j -= db + 1;
                } else {
                    data2[j--] = byte;
                }

                if (j < 0) {
                    break;
                }
            }
            memcpy(data, data2, sizeof(data));
        }
    }

    screen->br = data[0xff1] & 0xf;
    memset(screen->d[0], (data[0xff1] & 0xf) * 17, screen_w * screen_h);
    memset(screen->d[1], (data[0xff1] & 0xf) * 17, screen_w * screen_h);
    memset(data + 0xff2, data[0xff0], 6);

    ufli_pic(screen->d[0], data + 0x2000, data, data + 0x1000, data + 0xff2, data[0xff1] & 0xf);
    ufli_pic(screen->d[1], data + 0x6000, data + 0x4000, data + 0x5000, data + 0xff2, data[0xff1] & 0xf);
    return screen;
}

static view64_screen *PIC_shfxl(const unsigned char *mem, size_t size)
{
    unsigned char data[15360];
    unsigned char data2[0x2400];
    int x, y, x2, y2, i, j;
    unsigned char c0, c1, cs, b1, b2, *p, key, byte, db;
    const static unsigned char shfxlsprites[48] = {
        0x80, 0x84, 0x85, 0x89, 0x8A, 0x8E,
        0x8F, 0x93, 0x94, 0x98, 0x99, 0x9D,
        0x9E, 0xA2, 0xA3, 0xA7, 0xA8, 0xAC,
        0xAD, 0xB1, 0xB2, 0xB6, 0xB7, 0xBB,
        0xBC, 0xC0, 0xC1, 0xC5, 0xC6, 0xCA,
        0xCB, 0xCF, 0xD0, 0xD4, 0xD5, 0xD9,
        0xDA, 0xDE, 0xDF, 0xE3, 0xE4, 0xE8,
        0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE
    };
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 2) return NULL;
    screen = new_screen(screen_w, screen_h, 1);
    if (screen == NULL) return NULL;

    size -= 2;
    i = size;
    memcpy(data, mem + 2, (size < 15360) ? size : 15360);

    if (i != 15360 && i) {
        memset(data2, 0, sizeof(data2));
        j = 0x23cf;
        key = data[--i];
        if (i) {
            for (;; ) {
                byte = data[--i]; if (i == 0) {
                    break;
                }
                if (byte == key) {
                    db = data[--i] - 1; if (i == 0 || db > j) {
                        break;
                    }
                    byte = data[--i]; if (i == 0) {
                        break;
                    }
                    memset(&data2[j - db], byte, db + 1);
                    j -= db + 1;
                } else {
                    data2[j--] = byte;
                }

                if (j < 0) {
                    break;
                }
            }
        }

        memset(data, 255, sizeof(data));
        data[0x3e8] = data2[0xbd0];
        data[0x3e9] = data2[0xbd1];

        i = 0;
        for (y = 0; y < 168; y++) {
            for (x = 0; x < 18; x++)
            {
                data[shfxlsprites[((y - 1) & 0x7) * 6 + x / 3] * 64 + (x % 3) + (y % 21) * 3] = data2[i];
                data[0x2058 + (y & 0x7) + x * 8 + (y / 8) * 320] = data2[0xc00 + i];
                data[0x000b + (y & 0x7) * 1024 + x + (y / 8) * 40] = data2[0x1800 + i];
                i++;
            }
        }
    }

    screen->br = data[0x3e8] & 0xf;
    memset(screen->d[0], data[0x3e8] & 0xf, screen_w * screen_h);

    cs = data[0x3e9] & 0xf;

    for (y = 0; y < 21; y++) {
        for (x = 11; x < 29; x++)
        {
            p = screen->d[0] + (y * screen_w + x) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                c0 = data[y * 40 + x + y2 * 1024];
                c1 = c0 / 16;
                c0 &= 15;
                b1 = data[8192 + (y * 40 + x) * 8 + y2];
                b2 = data[shfxlsprites[((y2 - 1) & 7) * 6 + (x - 11) / 3] * 64 + ((y * 8 + y2) % 21) * 3 + ((x - 11) % 3)];
                for (x2 = 0; x2 < 8; x2++)
                {
                    *p = (b2 & 128) ? cs : ((b1 & 128) ? c1 : c0);
                    p++;
                    b1 *= 2;
                    b2 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
    return screen;
}

static void depack_sif(unsigned char *data, const unsigned char *data2, size_t i) {
    int j = 0x1fef;
    unsigned char key = data2[0], byte, db;

    if (i > 1) {
        for (;; ) {
            byte = data2[--i]; if (i == 1) {
                break;
            }
            if (byte == key) {
                db = data2[--i]; if (i == 1) {
                    break;
                }
                byte = data2[--i];
                if (db) {
                    memset(&data[j - db + 1], byte, db);
                    j -= db;
                } else {
                    memset(&data[j - 256 + 1], byte, 256);
                    j -= 256;
                }
            } else {
                data[j--] = byte;
            }

            if (i <= 1) {
                break;
            }
            if (j < 0) {
                break;
            }
        }
    }
}

static view64_screen *PIC_super_hires_interlace_fli(const unsigned char *mem, size_t size)
{
    unsigned char data[0x4000];
    int x, y, x2;
    unsigned char c0a, c0b, c1a, c1b, cs0a, cs0b, cs1a, cs1b, b1a, b1b, b2a, b2b, b3a, b3b, *p, *q;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;
    size_t i;

    if (size < 2) return NULL;
    screen = new_screen(screen_w, screen_h, 2);
    if (screen == NULL) return NULL;
    screen->br = 15;

    i = mem[0] + mem[1] * 256;
    if (i <= 0x9402 || i > 0xb402) {
        free(screen);
        return NULL;
    }
    i -= 0x9402;
    mem += 2; size -= 2;
    if (i > size) {
        free(screen);
        return NULL;
    }
    depack_sif(data, mem, i);
    size -= i;
    mem += i;

    if (size < 2) {
        free(screen);
        return NULL;
    }
    i = mem[0] + mem[1] * 256;
    if (i <= 0x9402 || i > 0xb402) {
        free(screen);
        return NULL;
    }
    i -= 0x9402;
    mem += 2; size -= 2;
    if (i > size) {
        free(screen);
        return NULL;
    }
    depack_sif(data + 0x2000, mem, i);
    size -= i;
    mem += i;

    if (size < 4) {
        free(screen);
        return NULL;
    }
    cs0a = mem[0] & 0xf;
    cs1a = mem[1] & 0xf;
    cs0b = mem[2] & 0xf;
    cs1b = mem[3] & 0xf;

    memset(screen->d[0], 15, screen_w * screen_h);
    memset(screen->d[1], 15, screen_w * screen_h);

    i = 0;
    for (y = 1; y < 168; y++) {
        for (x = 14; x < 26; x++)
        {
            p = screen->d[0] + (y * screen_w + x * 8);
            q = screen->d[1] + (y * screen_w + x * 8);
            c0a = data[0x1800 + i];
            c1a = c0a / 16;
            c0a &= 15;
            b1a = data[0x1000 + i];
            b2a = data[i];
            b3a = data[0x800 + i];

            c0b = data[0x3800 + i];
            c1b = c0b / 16;
            c0b &= 15;
            b1b = data[0x3000 + i];
            b2b = data[0x2000 + i];
            b3b = data[0x2800 + i];
            for (x2 = 0; x2 < 8; x2++)
            {
                p[x2] = (b2a & 128) ? cs0a : ((b3a & 128) ? cs1a : ((b1a & 128) ? c1a : c0a));
                q[x2] = (b2b & 128) ? cs0b : ((b3b & 128) ? cs1b : ((b1b & 128) ? c1b : c0b));
                b1a *= 2;
                b2a *= 2;
                b3a *= 2;
                b1b *= 2;
                b2b *= 2;
                b3b *= 2;
            }
            i++;
        }
    }
    return screen;
}

static void rlc_unpack(unsigned char *data, size_t max, const unsigned char *mem, size_t size)
{
    unsigned char byte, db;
    size_t pos = 0, p = 0;

    while (pos < max)
    {
        if (p >= size) break;
        db = mem[p++];
        if (db & 0x80) {
            if (p >= size) break;
            byte = mem[p++];
            db &= 0x7f;
            if (pos + db > max) db = max - pos;
            memset(data + pos, byte, db);
        } else {
            if (db > size) db = size;
            if (pos + db > max) db = max - pos;
            memcpy(data + pos, mem + p, db);
            p += db;
        }
        pos += db & 0x7f;
    }
}

static view64_screen *PIC_super_hires_interlace(const unsigned char *mem, size_t size)
{
    unsigned char data[15352 + 256];
    int x, y, x2, y2;
    unsigned char c0, c1, cs1, cs2, b1, b2, b3, b4, b5, b6, *p, *q;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 3) return NULL;
    screen = new_screen(screen_w, screen_h, 2);
    if (screen == NULL) return NULL;
    screen->br = 0;
    size -= 3;

    if (mem[2]) {
        rlc_unpack(data, 15352, mem + 3, size);
    } else {
        memcpy(data, mem + 3, (15352 < size) ? 15352 : size);
    }
    memset(screen->d[0], 0, screen_w * screen_h);
    memset(screen->d[1], 0, screen_w * screen_h);

    for (y = 0; y < 25; y++) {
        for (x = 0; x < 12; x++)
        {
            p = screen->d[0] + (y * screen_w + x + 14) * 8;
            q = screen->d[1] + (y * screen_w + x + 14) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                c0 = data[0x3ac0 + y * 12 + x];
                c1 = c0 / 16;
                c0 &= 15;
                b1 = data[(y * 12 + x) * 8 + y2];
                b2 = data[0x960 + (y * 12 + x) * 8 + y2];
                {
                    int yy = (x % 3) + ((y * 8 + y2) % 21) * 3;
                    int yy2 = ((y * 8 + y2) / 21) * 8 + (x / 3);
                    b3 = data[0x12c0 + yy + yy2 * 64];
                    b4 = data[0x12c0 + yy + (yy2 + 4) * 64];
                    b5 = data[0x26c0 + yy + yy2 * 64];
                    b6 = data[0x26c0 + yy + (yy2 + 4) * 64];
                    cs1 = data[0x3bec + x / 3] & 0xf;
                    cs2 = data[0x3bf0 + x / 3] & 0xf;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    p[x2] = (b3 & 128) ? cs1 : ((b4 & 128) ? cs2 : ((b1 & 128) ? c1 : c0));
                    q[x2] = (b5 & 128) ? cs1 : ((b6 & 128) ? cs2 : ((b2 & 128) ? c1 : c0));
                    b1 *= 2;
                    b2 *= 2;
                    b3 *= 2;
                    b4 *= 2;
                    b5 *= 2;
                    b6 *= 2;
                }
                p += screen_w;
                q += screen_w;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_extend_super_hires_interlace(const unsigned char *mem, size_t size)
{
    unsigned char data[20454 + 256];
    int x, y, x2, y2;
    unsigned char c0, c1, cs1, b1, b2, b3, b5, *p, *q;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen;

    if (size < 3) return NULL;
    screen = new_screen(screen_w, screen_h, 2);
    if (screen == NULL) return NULL;
    screen->br = 0;

    if (mem[2]) {
        rlc_unpack(data, 20454, mem + 3, size);
    } else {
        memcpy(data, mem + 3, (20454 < size) ? 20454 : size);
    }
    memset(screen->d[0], 0, screen_w * screen_h);
    memset(screen->d[1], 0, screen_w * screen_h);

    for (y = 0; y < 25; y++) {
        for (x = 0; x < 24; x++)
        {
            p = screen->d[0] + (y * screen_w + x + 8) * 8;
            q = screen->d[1] + (y * screen_w + x + 8) * 8;
            for (y2 = 0; y2 < 8; y2++)
            {
                c0 = data[0x4d80 + y * 24 + x];
                c1 = c0 / 16;
                c0 &= 15;
                b1 = data[(y * 24 + x) * 8 + y2];
                b2 = data[0x12c0 + (y * 24 + x) * 8 + y2];
                {
                    int yy = (x % 3) + ((y * 8 + y2) % 21) * 3;
                    int yy2 = ((y * 8 + y2) / 21) * 8 + (x / 3);
                    b3 = data[0x2580 + yy + yy2 * 64];
                    b5 = data[0x3980 + yy + yy2 * 64];
                    cs1 = data[0x4fd8 + x / 3] & 0xf;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    p[x2] = (b3 & 128) ? cs1 : ((b1 & 128) ? c1 : c0);
                    q[x2] = (b5 & 128) ? cs1 : ((b2 & 128) ? c1 : c0);
                    b1 *= 2;
                    b2 *= 2;
                    b3 *= 2;
                    b5 *= 2;
                }
                p += screen_w;
                q += screen_w;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_super_hires_studio(const unsigned char *mem)
{
    int x, y, x2, y2;
    unsigned char c[2], cs1, cs2, b1, b2, b3, *p;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = 0;

    for (x = 0; x < 40; x++)
    {
        int xx = x - 2;
        p = screen->d[0] + x * 8;
        if (xx >= 0 && xx < 12) {
            cs1 = mem[0x1f42 + xx / 3] & 15;
            cs2 = mem[0x1f46 + xx / 3] & 15;
        } else {
            cs1 = cs2 = 0;
        }

        for (y = 0; y < 25; y++) {
            c[0] = mem[0x3402 + y * 40 + x];
            c[1] = c[0] / 16;
            c[0] &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                int yy = y * 8 + y2 - 17;
                b1 = mem[2 + (y * 40 + x) * 8 + y2];
                if (yy >= 0 && yy < 167 && xx >= 0 && xx < 12) {
                    b2 = mem[0x2002 + (yy % 21) * 3 + xx % 3 + xx / 3 * 64 * 8 + yy / 21 * 64];
                    b3 = mem[0x2802 + (yy % 21) * 3 + xx % 3 + xx / 3 * 64 * 8 + yy / 21 * 64];
                } else {
                    b2 = b3 = 0;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    p[x2] = (b3 & 128) ? cs2 : ((b2 & 128) ? cs1 : c[b1 >> 7]);
                    b1 *= 2; b2 *= 2; b3 *= 2;
                }
                p += screen_w;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_super_hires(const unsigned char *mem)
{
    int x, y, x2, y2;
    unsigned char c0, c1, cs1, cs2, b1, b2, b3, *p;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = 15;

    memset(screen->d[0], 15, screen_w * screen_h);
    cs1 = mem[0xca6] & 15;
    cs2 = mem[0xca7] & 15;

    for (x = 0; x < 12; x++)
    {
        p = screen->d[0] + (x + 14) * 8;

        for (y = 0; y < 11; y++) {
            c0 = mem[0x422 + y * 12 + x];
            c1 = c0 / 16;
            c0 &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                int yy = y * 8 + y2;
                b1 = mem[2 + (y * 12 + x) * 8 + y2];
                if (yy < 84) {
                    b2 = mem[0x04a6 + (yy % 21) * 3 + x % 3 + x / 3 * 64 + yy / 21 * 64 * 8];
                    b3 = mem[0x05a6 + (yy % 21) * 3 + x % 3 + x / 3 * 64 + yy / 21 * 64 * 8];
                } else {
                    b2 = b3 = 0;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    *p = (b3 & 128) ? cs2 : ((b2 & 128) ? cs1 : ((b1 & 128) ? c1 : c0));
                    p++;
                    b1 *= 2;
                    b2 *= 2;
                    b3 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_super_hires_2(const unsigned char *mem)
{
    int x, y, x2, y2;
    unsigned char c0, c1, cs1, b1, b2, *p;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = 15;

    memset(screen->d[0], 15, screen_w * screen_h);

    for (x = 0; x < 24; x++)
    {
        p = screen->d[0] + x * 8;
        cs1 = mem[0x21ba + x / 3] & 15;
        for (y = 0; y < 21; y++) {
            c0 = mem[0x1fc2 + y * 24 + x];
            c1 = c0 / 16;
            c0 &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                int yy = y * 8 + y2 - 1;
                b1 = mem[2 + (y * 24 + x) * 8 + y2];
                if (yy >= 0) {
                    b2 = mem[0x0fc2 + (yy % 21) * 3 + x % 3 + x / 3 * 64 + yy / 21 * 64 * 8];
                } else {
                    b2 = 0;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    *p = (b2 & 128) ? cs1 : ((b1 & 128) ? c1 : c0);
                    p++;
                    b1 *= 2;
                    b2 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_super_hires_ii(const unsigned char *mem, size_t size)
{
    unsigned char data[0x2180];
    unsigned char key, db;
    int x, y, x2, y2;
    unsigned char c0, c1, cs1, b1, b2, *p;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);
    size_t i, j;

    if (screen == NULL) return NULL;
    screen->br = 0;

    if (size == 14770) {
        for (y = 0; y < 21; y++) {
            memcpy(data + 0x1f80 + y * 24, mem + 0x1ba + y * 40, 24);
            memcpy(data + y * 192, mem + 0x19f2 + y * 320, 192);
        }
        for (x = 0; x < 24; x++) {
            for (y = 0; y < 168; y++) {
                data[0xfc0 + y + x * 168] = mem[0x9b2 + (y % 21) * 3 + x % 3 + (x / 3) * 64 + (y / 21) * 64 * 8];
            }
        }
        memcpy(data + 0x2178, mem + 0x1aa, 8);
    } else {
        i = 1; j = 0;

        key = mem[0];
        while (i < size) {
            if (mem[i] == key) {
                i++;
                if (i >= size) {
                    break;
                }
                db = mem[i++];
                if (!db) {
                    break;
                }
                if (i >= size || j + db > sizeof(data)) {
                    break;
                }
                memset(data + j, mem[i], db);
                j += db;
                i++;
            } else {
                data[j++] = mem[i++];
            }
            if (j >= sizeof(data)) {
                break;
            }
        }
    }

    memset(screen->d[0], 0, screen_w * screen_h);

    for (x = 0; x < 24; x++)
    {
        p = screen->d[0] + (x + 8) * 8;
        cs1 = data[0x2178 + x / 3] & 15;
        for (y = 0; y < 21; y++) {
            c0 = data[0x1f80 + y * 24 + x];
            c1 = c0 / 16;
            c0 &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                b1 = data[(y * 24 + x) * 8 + y2];
                b2 = data[0x0fc0 + y * 8 + y2 + x * 168];
                for (x2 = 0; x2 < 8; x2++)
                {
                    *p = (b2 & 128) ? cs1 : ((b1 & 128) ? c1 : c0);
                    p++;
                    b1 *= 2;
                    b2 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_super_hires_i(const unsigned char *mem, size_t size)
{
    unsigned char data[0x18a0];
    unsigned char key, db;
    int x, y, x2, y2;
    size_t i, j;
    unsigned char c0, c1, cs1, cs2, b1, b2, b3, *p;
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->br = 0;
    //8000-a180

    if (size == 14770) {
        for (y = 0; y < 21; y++) {
            memcpy(data + 0x17a0 + y * 12, mem + 0x1c0 + y * 40, 12);
            memcpy(data + y * 96, mem + 0x1a22 + y * 320, 96);
        }
        for (x = 0; x < 12; x++) {
            for (y = 0; y < 168; y++) {
                data[0x7e0 + y + x * 168] = mem[0x9b2 + (y % 21) * 3 + x % 3 + (x / 3) * 64 + (y / 21) * 64 * 4];
                data[0xfc0 + y + x * 168] = mem[0x11b2 + (y % 21) * 3 + x % 3 + (x / 3) * 64 + (y / 21) * 64 * 4];
            }
        }

        for (x = 0; x < 4; x++) {
            data[0x189c + x] = (mem[0x1ae + x] << 4) | (mem[0x1aa + x] & 15);
        }
    } else {
        i = 1; j = 0;

        key = mem[0];
        while (i < size) {
            if (mem[i] == key) {
                i++;
                if (i >= size) {
                    break;
                }
                db = mem[i++];
                if (!db) {
                    break;
                }
                if (i >= size || j + db > sizeof(data)) {
                    break;
                }
                memset(data + j, mem[i], db);
                j += db;
                i++;
            } else {
                data[j++] = mem[i++];
            }
            if (j >= sizeof(data)) {
                break;
            }
        }
    }

    memset(screen->d[0], 0, screen_w * screen_h);

    for (x = 0; x < 12; x++)
    {
        p = screen->d[0] + (x + 14) * 8;
        cs1 = data[0x189c + x / 3];
        cs2 = cs1 >> 4;
        cs1 &= 15;

        for (y = 0; y < 21; y++) {
            c0 = data[0x17a0 + y * 12 + x];
            c1 = c0 / 16;
            c0 &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                b1 = data[(y * 12 + x) * 8 + y2];
                b2 = data[0x07e0 + y * 8 + y2 + x * 168];
                b3 = data[0x0fc0 + y * 8 + y2 + x * 168];
                for (x2 = 0; x2 < 8; x2++)
                {
                    *p = (b3 & 128) ? cs2 : ((b2 & 128) ? cs1 : ((b1 & 128) ? c1 : c0));
                    p++;
                    b1 *= 2;
                    b2 *= 2;
                    b3 *= 2;
                }
                p += screen_w - 8;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_fuckpaint(const unsigned char *mem)
{
    return mci_pic(mem + 0xc02, mem + 0x2c02, mem + 0x402, mem + 0x802, mem + 0x2, mem[0x2b42], mem[0x2b42], 0);
}

static view64_screen *PIC_truepaint(const unsigned char *mem, size_t size)
{
    unsigned char data2[19432];
    unsigned char a, x;
    int i, j, k;

    i = size;

    if (i != 19434 && i > 2) {
        i--;
        j = 19431;
        while (i > 2 && j >= 0) {
            a = mem[i--]; x = 1;
            for (k = 8; k >= 0; k--) {
                if (a == mem[127 + k + 2]) {
                    switch (mem[118 + k + 2]) {
                    case 0x00: x++; a = mem[131 + k + 2]; break;
                    case 0x07: x = mem[i--]; if (i < 2) {
                            break;
                    }
                    case 0x0b: x += 2; a = mem[77 + 2]; break;
                    case 0x11: x = mem[i--]; if (i < 2) {
                            break;
                    }
                    case 0x17: x += 2;
                    case 0x19: a = mem[i--]; break;
                    }
                    break;
                }
            }
            for (;; ) {
                data2[j--] = a;
                x--;
                if (!x || j < 0) {
                    break;
                }
            }
        }
    } else {
        memcpy(data2, mem + 2, sizeof(data2));
    }

    return mci_pic(data2 + 0x400, data2 + 0x2400, data2, data2 + 0x4400, data2 + 0x4800, data2[1000], 0, 0);
}

static view64_screen *PIC_gunpaint(const unsigned char *mem)
{
    unsigned char bgcolors[200];

    memcpy(bgcolors, mem + 0x3f42, 177);
    memcpy(bgcolors + 177, mem + 0x47ea, 23);
    return ifli_pic(mem + 0x2002, mem + 0x6402, mem + 0x0002, mem + 0x4402, mem + 0x4002, bgcolors);
}

/* start: $1800, len: 10050, name: PICASSO 64 / FLYING COLORS
 * $1800 - color 2
 * $1c00 - color 1
 * $1ffe - border
 * $1fff - background
 * $2000 - bitmap
 *
 * $1ffd - must be 0, $1ff9-$1ffc reserved
 */
static view64_screen *PIC_picasso64(const unsigned char *mem)
{
    return multi_pic(mem + 0x0802, mem + 0x0402, mem + 0x0002, mem[0x0800], mem[0x0801]);
}

static view64_screen *PIC_wigmore_artist(const unsigned char *mem)
{
    return multi_pic(mem + 0x0002, mem + 0x2002, mem + 0x2402, mem[0x2800], mem[0x2801]);
}

static view64_screen *PIC_art_studio_multi(const unsigned char *mem)
{
    return multi_pic(mem + 0x0002, mem + 0x1f42, mem + 0x233a, mem[0x232a], mem[0x232b]);
}

static view64_screen *PIC_art_studio_multi_window(const unsigned char *mem, size_t size)
{
    int x, y, x2, y2, i;
    unsigned char c[4], b, *p;
    int w, h, ox, oy;
    view64_screen *screen;
    
    if (size < 5) return NULL;
    ox = (mem[1] & 3) * 2;
    oy = mem[2] & 7;
    w = mem[3] * 2;
    h = mem[4];

    screen = new_screen(w, h, 1);
    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = 0;

    i = 5;
    for (y = 0; y < h + oy; y += 8) {
        for (x = 0; x < w + ox; x += 8) {
            c[2] = (i < size) ? mem[i++] : 0;
            c[1] = c[2] / 16;
            c[2] &= 15;
            c[3] = (i < size) ? (mem[i++] & 15) : 0;
            p = screen->d[0] + (y - oy) * w + x - ox;
            for (y2 = 0; y2 < 8; y2++)
            {
                b = i < size ? mem[i++] : 0;
                if (y2 + y >= oy && y + y2 - oy < h) {
                    c[0] = (y2 & 1) ? 11 : 12;
                    for (x2 = 0; x2 < 8; x2 += 2) {
                        if (x2 + x >= ox && x + x2 - ox < w) {
                            unsigned char ki = b >> 6;
                            c[0] ^= 11^12;
                            p[x2] = c[ki];
                            c[0] ^= 11^12;
                            p[x2 + 1] = c[ki];
                            b <<= 2;
                        }
                    }
                }
                p += w;
            }
        }
    }
    return screen;
}

static view64_screen *PIC_zoomatic(const unsigned char *mem, size_t size)
{
    unsigned char data[0x2711];
    unsigned char key, byte, db;
    int i, j;

    i = size;

    j = 0x2710;
    if (i > 2) {
        key = mem[--i];
        for (;; ) {
            if (i <= 2) {
                break;
            }
            byte = mem[--i];
            if (byte == key) {
                if (i <= 2) {
                    break;
                }
                db = mem[--i];
                if (i <= 2) {
                    break;
                }
                byte = mem[--i];
                if (db) {
                    memset(&data[j - db + 1], byte, db);
                    j -= db;
                } else {
                    memset(&data[j - 256 + 1], byte, 256);
                    j -= 256;
                }
            } else {
                data[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }
    }
    return multi_pic(data, data + 8000, data + 9000, data[0x2710] >> 4, data[0x2710] & 0xf);
}

static view64_screen *PIC_drazpaint(const unsigned char *mem, size_t size)
{
    unsigned char data[10049 + 256];

    if (memcmp(mem + 2, "DRAZPAINT 1.3", 13) &&
        memcmp(mem + 2, "DRAZPAINT 1.4", 13) &&
        memcmp(mem + 2, "DRAZPAINT 2.0", 13)) {
        memcpy(data, mem + 2, 10049);
    } else {
        rle_depack(mem + 16, size - 16, mem[15], data, 10049, RLE_BYTE_LEN);
    }
    return multi_pic(data + 2048, data + 1024, data, data[10048], data[10048]);
}

static view64_screen *PIC_drazlace(const unsigned char *mem, size_t size)
{
    unsigned char data[18240 + 256];

    if (memcmp(mem + 2, "DRAZLACE! 1.0", 13)) {
        memcpy(data, mem + 2, 18240);
    } else { 
        rle_depack(mem + 16, size - 16, mem[15], data, 18240, RLE_BYTE_LEN);
    }
    return mci_pic(data + 2048, data + 10240, data + 1024, data + 1024, data, data[10048], data[10048], !data[10050]);
}

static view64_screen *PIC_funpaint(const unsigned char *mem, size_t size) //.fun
{
    unsigned char data[33676 + 256], bgcolors[200];

    if (mem[16] && !memcmp(mem + 2, "FUNPAINT (MT) ", 14)) {
        rle_depack(mem + 18, size - 18, mem[17], data, 33676, RLE_BYTE_LEN);
    } else {
        memcpy(data, mem + 18, 33676);
    }

    memcpy(bgcolors, data + 16200, 100);
    memcpy(bgcolors + 100, data + 33576, 100);
    return ifli_pic(data + 8192, data + 25576, data, data + 17384, data + 16384, bgcolors);
}

static view64_screen *PIC_pixelperfect(const unsigned char *mem, size_t size) //.pp .ppp
{
    unsigned char data[33600 + 256], bgcolors[200];
    int i;
    int y;
    unsigned char profi[0x480];
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 2);

    if (screen == NULL) return NULL;
    screen->br = 0;
    screen->flags |= VIEW64_SCREEN_MCI;

    if (mem[2] == 0x10 && mem[3] == 0x10 && mem[4] == 0x10) {
        rle_depack(mem + 6, size - 6, mem[5], data, 33600, RLE_BYTE_LEN1);
    } else {
        memcpy(data, mem + 2, 33600);
    }

    for (i = 0; i < 25; i++) {
        memset(data + 0x2400 + i * 320, 255, 24);
        memset(data + 0x6400 + i * 320, 255, 24);
    }
    memset(bgcolors, data[0x437f], 200);
    memset(profi, 0, 0x480);
    fli_pic2(screen->d[0], data + 0x2400, data + 0x400, data, bgcolors, profi);
    fli_pic2(screen->d[1] + 1, data + 0x6400, data + 0x4400, data, bgcolors, profi);
    for (y = 0; y < screen_h; y++) screen->d[1][y * screen_w] = 0;
    return screen;
}

static view64_screen *PIC_facepaint(const unsigned char *mem)
{
    return multi_pic(mem + 2, mem + 8002, mem + 9002, mem[10003], mem[10002]);
}

static view64_screen *PIC_art_studio_hires(const unsigned char *mem, size_t size)
{
    unsigned char data[9001];

    if (size < 2) return NULL;
    size -= 2;
    
    memset(data, 0, 8000);
    memset(data + 8000, 0x10, 1000);
    data[9000] = 0;

    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));
    return hires_pic(data, data + 8000, data[9000]);
}

static view64_screen *PIC_font(const unsigned char *mem, size_t size)
{
    int x;
    unsigned char bitmap[8000], color[1000];

    if (size < 2) return NULL;
    size -= 2;
    memset(bitmap, 0, sizeof(bitmap));
    memset(color, 0x10, sizeof(color));

    for (x = 0; x < 8; x++) {
        memcpy(bitmap + x * 320, mem + 2, (size < 256) ? size : 256);
        if (size < 256) break;
        size -= 256;
        mem += 256;
    }
    return hires_pic(bitmap, color, 0);
}

static view64_screen *hires_interlace_pic(const unsigned char *bmp, const unsigned char *col, const unsigned char *bmp2, const unsigned char *col2)
{
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 2);

    if (screen == NULL) return NULL;
    screen->br = 0;

    hires_pic2(screen->d[0], bmp, col);
    hires_pic2(screen->d[1], bmp2, col2);
    return screen;
}

static view64_screen *PIC_hires_interlace(const unsigned char *mem)
{
    return hires_interlace_pic(mem + 0x0002, mem + 0x2802, mem + 0x4002, mem + 0x2402);
}

static view64_screen *PIC_hireslace_editor(const unsigned char *mem)
{
    return hires_interlace_pic(mem + 0x0002, mem + 0x2002, mem + 0x6002, mem + 0x4002);
}

static view64_screen *PIC_doodle(const unsigned char *mem, size_t size)
{
    unsigned char data[9216 + 256];

    if (size == 9218) {
        memcpy(data, mem + 2, 9216);
    } else {
        rle_depack(mem + 2, size - 2, 0xfe, data, 9216, RLE_LEN_BYTE);
    }
    return hires_pic(data + 1024, data, 0);
}

static view64_screen *PIC_image_system_hires(const unsigned char *mem)
{
    return hires_pic(mem + 2, mem + 8194, 0); //no border
}

static view64_screen *PIC_hieddi(const unsigned char *mem, size_t size)
{
    unsigned char data[9192];

    if (size < 2) return NULL;
    size -= 2;
    memset(data, 0, 8000);
    memset(data + 0x2000, 0x63, 1000);
    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));

    return hires_pic(data, data + 0x2000, 0);
}

static view64_screen *PIC_hires_editor(const unsigned char *mem)
{
    return hires_pic(mem + 1026, mem + 2, 0);
}

static view64_screen *PIC_image_system_multi(const unsigned char *mem)
{
    return multi_pic(mem + 1026, mem + 9218, mem + 2, 0, mem[9217]); //border nincs
}

static view64_screen *PIC_centauri_logo_editor(const unsigned char *mem)
{
    unsigned char chr[1000], col[1000];
    memset(chr, mem[8002], sizeof(chr));
    memset(col, mem[8003], sizeof(col));
    return multi_pic(mem + 2, chr, col, 0, mem[8004]); //border nincs
}

static view64_screen *PIC_interlaced_logo_editor(const unsigned char *mem)
{
    unsigned char chr[1000], col[1000], bmp[8000], bmp2[8000];
    memset(chr, (mem[4095] << 4) | (mem[4096] & 15), sizeof(chr));
    memset(col, mem[4097] & 7, sizeof(col));
    memcpy(bmp, mem + 2050, 1920);
    memset(bmp + 1920, 0, 6080);
    memcpy(bmp2, mem + 2, 1920);
    memset(bmp2 + 1920, 0, 6080);

    return mci_pic(bmp, bmp2, chr, chr, col, mem[4094], 0, 0); //fixed border
}


static view64_screen *PIC_vidcom64(const unsigned char *mem)
{
    return multi_pic(mem + 0x802, mem + 0x402, mem + 0x2, 0, mem[0x8ea]);
}

static view64_screen *PIC_cdupaint(const unsigned char *mem)
{
    mem += 275;
    return multi_pic(mem, mem + 8000, mem + 9000, mem[10001], mem[10000]);
}

static view64_screen *PIC_koala_painter(const unsigned char *mem, size_t size)
{
    unsigned char data[10001 + 256];

    if (size >= 10002 && size <= 10003) {
        data[10000] = 0; // sometimes the background is missed
        memcpy(data, mem + 2, size - 2);
    } else {
        rle_depack(mem + 2, size - 2, 0xfe, data, 10001, RLE_LEN_BYTE);
    }
    return multi_pic(data, data + 8000, data + 9000, data[10000] / 16, data[10000] & 0x0f);
}

static view64_screen *PIC_blazing_paddles(const unsigned char *mem)
{
    return multi_pic(mem + 0x2, mem + 0x2002, mem + 0x2402, mem[0x1f81], mem[0x1f82]);
}

static view64_screen *PIC_artist64(const unsigned char *mem)
{
    return multi_pic(mem + 0x2, mem + 0x2002, mem + 0x2402, mem[0x2800], mem[0x2801]);
}

static view64_screen *PIC_cheese_paint(const unsigned char *mem)
{
    return multi_pic(mem + 0x2, mem + 0x4202, mem + 0x4802, 7, mem[0x4fff]);
}

static view64_screen *PIC_garfield(const unsigned char *mem)
{
    return multi_pic(mem + 2, mem + 8002, mem + 9002, 2, mem[10002]);
}

/* start: $7800, len: 10219, name: SARACEN PAINT
 * $7800 - color 1
 * $7bf0 - background
 * $7c00 - bitmap
 * $9c00 - color 2
 * border fixed to 12
 */
static view64_screen *PIC_saracen_paint(const unsigned char *mem)
{
    return multi_pic(mem + 0x402, mem + 0x2, mem + 0x2402, 12, mem[0x3f2]);
}

static view64_screen *PIC_dolphin(const unsigned char *mem)
{
    return multi_pic(mem + 0x802, mem + 0x402, mem + 0x2, 0, mem[0x7ea]);
}

static view64_screen *PIC_micro_illustrator(const unsigned char *mem, size_t size)
{
    unsigned char data[10020], data2[10020 + 256];
    size_t i, j, i2, l;

    if (size < 2) return NULL;
    size -= 2;
    i = size;
    memcpy(data, mem + 2, (size < sizeof(data)) ? size : sizeof(data));
    if (data[5] && i > 5) {
        i2 = 20;
        j = 20;
        while (i2 < i - 1 && j < sizeof(data2) - 256) {
            if (data[i2] == 0) {
                l = data[i2 + 1] * 256 + data[i2 + 2];
                if (j + l > sizeof(data2) - 256) {
                    l = sizeof(data2) - 256 - j;
                }
                memset(data2 + j, data[i2 + 3], l);
                j += l;
                i2 += 4;
            } else
            if (data[i2] < 128) {
                memset(data2 + j, data[i2 + 1], data[i2]); j += data[i2]; i2 += 2;
            } else {
                memcpy(data2 + j, data + i2 + 1, data[i2] - 128); j += data[i2] - 128; i2 = i2 + 1 + data[i2] - 128;
            }
        }
        switch (data[5]) {
        case 1:
            for (j = 0; j < 40; j++) {
                for (i = 0; i < 13; i++) {
                    data[20 + i * 80 + j] = data2[20 + j * 25 + i];
                    data[1020 + i * 80 + j] = data2[1020 + j * 25 + i];
                    if (i != 12) {
                        data[20 + 40 + i * 80 + j] = data2[20 + j * 25 + i + 13];
                        data[1020 + 40 + i * 80 + j] = data2[1020 + j * 25 + i + 13];
                    }
                }
            }
            for (j = 0; j < 40; j++) {
                for (i = 0; i < 100; i++) {
                    data[2020 + j * 8 + (i & 3) * 2 + (i & 0xfc) * 80] = data2[2020 + j * 200 + i];
                    data[2021 + j * 8 + (i & 3) * 2 + (i & 0xfc) * 80] = data2[2020 + j * 200 + i + 100];
                }
            }
            break;
        case 2:
            memcpy(data + 20, data2 + 20, 2000);
            for (i = 0; i < 200; i++) {
                for (j = 0; j < 40; j++) {
                    data[2020 + j * 8 + (i & 7) + (i & 0xf8) * 40] = data2[2020 + j + i * 40];
                }
            }
            break;
        }
    }
    return multi_pic(data + 2020, data + 20, data + 1020, data[6] >> 4, data[6]);
}

static view64_screen *PIC_afli(const unsigned char *mem) //.afl
{
    return iafli_pic(mem + 0x2002, mem + 0x002, mem + 0x2002, mem + 0x002);
}

static view64_screen *PIC_flimatic(const unsigned char *mem, size_t size)
{
    unsigned char data[0x4400];
    unsigned char backg[200], profi[0x480];
    unsigned char key, byte, db;
    int i, j;

    i = size;

    j = 0x437f;
    if (i == 0x4402) {
        memcpy(data, mem + 2, sizeof(data));
    } else
    if (i) {
        key = mem[--i];
        for (;; ) {
            if (!i) {
                break;
            }
            byte = mem[--i];
            if (byte == key) {
                if (!i) {
                    break;
                }
                db = mem[--i];
                if (!i) {
                    break;
                }
                byte = mem[--i];
                if (db) {
                    memset(&data[j - db + 1], byte, db);
                    j -= db;
                } else {
                    memset(&data[j - 256 + 1], byte, 256);
                    j -= 256;
                }
            } else {
                data[j--] = byte;
            }

            if (j < 0) {
                break;
            }
        }
    }

    memset(profi, 0, 0x480);
    memset(backg, data[0x437f], 200);
    for (i = 0; i < 1000; i++) {
        if (data[i]) {
            return fli_pic(data + 0x2400, data + 0x400, data, backg, profi, 0);
        }
    }
    return iafli_pic(data + 0x2400, data + 0x400, data + 0x2400, data + 0x400);
}

static view64_screen *PIC_hires_manager(const unsigned char *mem, size_t size)
{
    unsigned char data[0x3fff];
    int i, j, l;

    if (size < 2) return NULL;
    i = size - 2;
    mem += 2;

    if (i == 0x3fff) {
        memcpy(data, mem, sizeof(data));
    } else {
        j = 0x3ff1;
        i--;
        while (i >= 0) {
            if (mem[i]) {
                l = mem[i];
                i -= l; j -= l - 1;
                if (i < 0 || j < 0) {
                    break;
                }
                memcpy(data + j, mem + i, l);
            } else {
                i--;
                if (i < 0) {
                    break;
                }
                l = mem[i];
                j -= l;
                i--;
                if (j < 0 || i < 0) {
                    break;
                }
                memset(data + j, mem[i], l + 1);
                i--;
            }
        }
        //data[0x3ffe]=mem[4];
        //data[0x3ff0]=3;
    }
    memset(data, 255, 64);
    memset(data + 64, 0, 256);
    memset(data + 0x2000, 0, 40);

    return iafli_pic(data, data + 0x2000, data, data + 0x2000);
}

static view64_screen *PIC_hires_fli_design(const unsigned char *mem)
{
    return iafli_pic(mem + 2, mem + 8194, mem + 2, mem + 8194);
}

static view64_screen *PIC_eci(const unsigned char *mem, size_t size)
{
    unsigned char data[32768 + 256];

    if (size == 32770) {
        memcpy(data, mem + 2, 32768);
    } else {
        rle_depack(mem + 3, size - 3, mem[2], data, 32768, RLE_BYTE_LEN);
    }

    return iafli_pic(data, data + 8192, data + 16384, data + 24576);
}

static view64_screen *PIC_amica(const unsigned char *mem, size_t size)
{
    unsigned char data[10001 + 256];

    rle_depack(mem + 2, size - 2, 0xc2, data, 10001, RLE_BYTE_LEN);
    return multi_pic(data, data + 8000, data + 9000, 0, data[10000]);
}

static view64_screen *PIC_paint_magic(const unsigned char *mem)
{
    unsigned char color2[1000];

    memset(color2, mem[0x1fb7], 1000);
    return multi_pic(mem + 0x74, mem + 0x2074, color2, mem[0x1fb8], mem[0x1fb4]);
}

static view64_screen *PIC_paint_magic_plus(const unsigned char *mem)
{
    return multi_pic(mem + 0x74, mem + 0x2074, mem + 0x2474, mem[0x1fb8], mem[0x1fb4]);
}

static view64_screen *PIC_bitmap_editor(const unsigned char *mem) //.bed
{
    return multi_pic(mem + 0x802, mem + 0x402, mem + 0x002, 0, 0);
}

static view64_screen *PIC_rainbow_painter(const unsigned char *mem)
{
    return multi_pic(mem + 0x0402, mem + 0x0002, mem + 0x2402, 0, 0);
}

static view64_screen *PIC_run_paint_multi(const unsigned char *mem)
{
    return multi_pic(mem + 2, mem + 8002, mem + 9002, 0, mem[10002]);
}

static view64_screen *PIC_run_paint_hires(const unsigned char *mem)
{
    return hires_pic(mem + 1026, mem + 2, 0);
}

static view64_screen *PIC_run_paint_mono(const unsigned char *mem)
{
    unsigned char data[1000];
    memset(data, 1, sizeof(data));
    return hires_pic(mem + 2, data, 0);
}

static void super_hires(unsigned char *d, const unsigned char *mem)
{
    int x, y, x2, y2;
    unsigned char c[2], cs1, cs2, b1, b2, b3, *p;
    const int screen_w = 320;

    for (x = 0; x < 40; x++)
    {
        p = d + x * 8;
        if (x < 12) {
            cs1 = mem[0x3c02 + x / 3] & 15;
            cs2 = mem[0x3c06 + x / 3] & 15;
        } else {
            cs1 = cs2 = 0;
        }

        for (y = 0; y < 25; y++) {
            c[0] = mem[0x3402 + y * 40 + x];
            c[1] = c[0] / 16;
            c[0] &= 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                int yy = y * 8 + y2;
                b1 = mem[2 + (y * 40 + x) * 8 + y2];
                if (x < 12) {
                    b2 = mem[0x2002 + (yy % 21) * 3 + x % 3 + x / 3 * 64 * 10 + yy / 21 * 64];
                    b3 = mem[0x2a02 + (yy % 21) * 3 + x % 3 + x / 3 * 64 * 10 + yy / 21 * 64];
                } else {
                    b2 = b3 = 0;
                }
                for (x2 = 0; x2 < 8; x2++)
                {
                    p[x2] = (b3 & 128) ? cs2 : ((b2 & 128) ? cs1 : c[b1 >> 7]);
                    b1 *= 2;
                    b2 *= 2;
                    b3 *= 2;
                }
                p += screen_w;
            }
        }
    }
}

static view64_screen *PIC_interlace_super_hires(const unsigned char *mem)
{
    const int screen_w = 320, screen_h = 200;
    view64_screen *screen = new_screen(screen_w, screen_h, 2);

    if (screen == NULL) return NULL;
    screen->br = 0;

    super_hires(screen->d[0], mem);
    super_hires(screen->d[1], mem + 0x3c08);
    return screen;
}

/* start: $1800, len: 4098/4174, name: Logo-Painter V3/Logo Painter V3+
 * $1800-$1ff7 screen
 * $1ffb       $d021 (V3+ only)
 * $1ffc       $d800 (V3+ only)
 * $1ffd       $d022 (V3+ only)
 * $1ffe       $d023 (V3+ only)
 * $2000-$27ff font
 * $2800-      display code (V3+ only)
 * fixed black border
 */
static view64_screen *PIC_logo_painter(const unsigned char *mem, size_t size)
{
    int x, y, x2, y2;
    unsigned char c[4], b, *p;
    const int screen_w = 320, screen_h = 408;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);
    int multi;

    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = 0;

    if (size == 4174) {
        if (mem[0x7fe] & 8) {
            c[0] = mem[0x7fd] & 15;
            c[1] = mem[0x7ff] & 15;
            c[2] = mem[0x800] & 15;
            c[3] = mem[0x7fe] & 7;
            multi = 1;
        } else {
            c[0] = mem[0x7fd] & 15;
            c[1] = mem[0x7fe] & 7;
            multi = 0;
        }
    } else {
        c[0] = 0;
        c[1] = 10;
        c[2] = 2;
        c[3] = 1;
        multi = 1;
    }

    p = screen->d[0];
    for (y = 0; y < 51; y++) {
        for (x = 0; x < 40; x++)
        {
            for (y2 = 0; y2 < 8; y2++)
            {
                b = mem[0x802 + mem[0x2 + x + y * 40] * 8 + y2];
                if (multi) {
                    for (x2 = 0; x2 < 8; x2 += 2)
                    {
                        unsigned char ki = c[b >> 6];
                        p[x2] = ki;
                        p[x2 + 1] = ki;
                        b <<= 2;
                    }
                } else {
                    for (x2 = 0; x2 < 8; x2++)
                    {
                        p[x2] = c[b >> 7];
                        b <<= 1;
                    }
                }
                p += screen_w;
            }
            p -= screen_w * 8 - 8;
        }
        p += screen_w * 7;
    }
    return screen;
}

static view64_screen *PIC_bigfli(const unsigned char *mem)
{
    int x, y, x2, y2, cp;
    unsigned char c[4], b, *p;
    const unsigned char *color2, *colors, *bitmap;
    const int screen_w = 320, screen_h = 400;
    view64_screen *screen = new_screen(screen_w, screen_h, 1);

    if (screen == NULL) return NULL;
    screen->flags |= VIEW64_SCREEN_BORDERLESS;
    screen->br = 0;

    color2 = mem + 3;
    colors = mem + 0x403;
    bitmap = mem + 0x2403;

    c[0] = 0;
    cp = 0;
    p = screen->d[0];
    for (y = 0; y < 50; y++) {
        for (x = 0; x < 40; x++)
        {
            c[3] = (x < 3) ? 9 : color2[cp] & 15;
            for (y2 = 0; y2 < 8; y2++)
            {
                if (y < 25) {
                    c[2] = (x < 3) ? 0xff : colors[cp + y2 * 1024];
                    b = bitmap[cp * 8 + y2];
                } else {
                    c[2] = (x < 3) ? 0xff : colors[0x4000 + cp + y2 * 1024];
                    b = bitmap[0x4000 + cp * 8 + y2];
                }
                c[1] = c[2] / 16;
                c[2] &= 15;
                for (x2 = 0; x2 < 8; x2 += 2)
                {
                    unsigned char ki = c[b >> 6];
                    p[x2] = ki;
                    p[x2 + 1] = ki;
                    b *= 4;
                }
                p += screen_w;
            }
            p -= screen_w * 8 - 8;
            cp = (cp + 1) & 1023;
        }
        p += screen_w * 7;
    }
    return screen;
}

static const char *view64_format_names[] = {
    "Unknown",
    "4 bit GoDot",
    "VBM bitmap v2/v3",
    "Fli designer v2.0",
    "EMC-editor",
    "FLI editor/Saga",
    "Fli Graph v2.2",
    "Flip v1.33",
    "Fli Profi v1.0",
    "Flash Fli",
    "HCB-editor v0.05",
    "Super Hires Fli editor v1.0",
    "X-FLI editor v2.1",
    "UFLI-editor v1.0/v2.0",
    "MUFLI editor v1.0",
    "NUFLI editor v1.0",
    "MUIFLI editor v0.2",
    "UIFLI-editor v1.0",
    "SHF-XL edit v1.0",
    "Super Hires Interlace Fli editor v1.0",
    "Super Hires Interlace editor v1.3",
    "Extend Super Hires Interlace editor",
    "Super hires studio",
    "Super hires editor",
    "Super hires editor 2",
    "Super-hires editor II",
    "Super-hires editor I",
    "?",
    "Truepaint I",
    "Gunpaint IFLI",
    "Wigmore artist",
    "Art Studio v1.2b",
    "Zoomatic",
    "Drazpaint v1.3/1.5/2.0",
    "Draz-lace v1.0",
    "Funpaint II",
    "Pixel perfect",
    "Face-painter v1.0",
    "Art Studio v1.1",
    "Font",
    "Hires interlace editor v1.0",
    "Hireslace editor v1.5",
    "Doodle paint",
    "Image System",
    "Hi-eddi",
    "Hires-editor v1.0",
    "Image System",
    "Vidcom 64",
    "CDU Paint",
    "Koala painter",
    "Blazing Paddles",
    "Artist 64",
    "Cheese paint V2",
    "Create with Garfield",
    "Saracen paint",
    "Dolphin Ed",
    "Micro illustrator",
    "AFLI-editor v2.1",
    "Flimatic v3.7",
    "Hires manager v1.0",
    "Hires FLI Design",
    "ECIgraphic-editor v1.0",
    "Amica paint 1.0",
    "Paint magic",
    "Bitmap editor",
    "Rainbow painter",
    "Run paint",
    "Run paint",
    "Run paint",
    "Interpaint V1.0+",
    "Interpaint V1.0+",
    "Giga-Paint",
    "Giga-Paint",
    "FLI designer v2.0",
    "Interlace-Super-Hires Painter v1.0",
    "Picasso 64",
    "Logo-Painter V3",
    "Big FLI",
    "Flying colors",
    "Centauri logo-editor V2.1",
    "Interlaced logo editor V1.01",
    "Art studio v1.2b window",
    "Paint Magic+",
    "Charpad",
    "Spritepad"
};

const struct {
    const char *ext;
    view64_format format;
} view64_extensions[] = {
    {"fp",   VIEW64_FUCKPAINT},
    {"pp",   VIEW64_PIXELPERFECT},
    {"jj",   VIEW64_DOODLE},
    {"dd",   VIEW64_DOODLE},
    {"rp",   VIEW64_RAINBOW_PAINTER},
    {"art",  VIEW64_ART_STUDIO_HIRES}, 
    {"hpi",  VIEW64_ART_STUDIO_HIRES}, 
    {"hir",  VIEW64_ART_STUDIO_HIRES}, 
    {"iph",  VIEW64_INTERPAINT_HIRES}, 
    {"hre",  VIEW64_INTERPAINT_HIRES}, 
    {"ocp",  VIEW64_ART_STUDIO_MULTI}, 
    {"mpi",  VIEW64_ART_STUDIO_MULTI}, 
    {"mpic", VIEW64_ART_STUDIO_MULTI}, 
    {"mwi",  VIEW64_ART_STUDIO_MULTI_WINDOW}, 
    {"mwin", VIEW64_ART_STUDIO_MULTI_WINDOW}, 
    {"drl",  VIEW64_DRAZLACE}, 
    {"gun",  VIEW64_GUNPAINT}, 
    {"ifl",  VIEW64_GUNPAINT}, 
    {"him",  VIEW64_HIRES_MANAGER},
    {"afl",  VIEW64_AFLI},
    {"ecp",  VIEW64_ECI},
    {"eci",  VIEW64_ECI},
    {"fli",  VIEW64_FLI},
    {"fp2",  VIEW64_FUNPAINT},
    {"fun",  VIEW64_FUNPAINT},
    {"ppp",  VIEW64_PIXELPERFECT},
    {"hcb",  VIEW64_HCB},
    {"drz",  VIEW64_DRAZPAINT},
    {"drp",  VIEW64_DRAZPAINT},
    {"kla",  VIEW64_KOALA_PAINTER},
    {"koa",  VIEW64_KOALA_PAINTER},
    {"ipt",  VIEW64_INTERPAINT_MULTI},
    {"lre",  VIEW64_INTERPAINT_MULTI},
    {"fpt",  VIEW64_FACEPAINT},
    {"fcp",  VIEW64_FACEPAINT},
    {"ami",  VIEW64_AMICA},
    {"bpl",  VIEW64_BLAZING_PADDLES},
    {"shf",  VIEW64_SUPER_HIRES_FLI},
    {"shi",  VIEW64_SUPER_HIRES_INTERLACE},
    {"ufl",  VIEW64_UFLI},
    {"shx",  VIEW64_SHFXL},
    {"sif",  VIEW64_SUPER_HIRES_INTERLACE_FLI},
    {"ism",  VIEW64_IMAGE_SYSTEM_MULTI},
    {"mci",  VIEW64_TRUEPAINT},
    {"cdu",  VIEW64_CDUPAINT},
    {"hfc",  VIEW64_HIRES_FLI_DESIGN},
    {"hfd",  VIEW64_HIRES_FLI_DESIGN},
    {"ddp",  VIEW64_DOODLE},
    {"vid",  VIEW64_VIDCOM64},
    {"ffl",  VIEW64_FLASH_FLI},
    {"ffli", VIEW64_FLASH_FLI},
    {"xfl",  VIEW64_XFLI},
    {"uif",  VIEW64_UIFLI},
    {"mup",  VIEW64_MUFLI},
    {"muf",  VIEW64_MUFLI},
    {"fpr",  VIEW64_FLIPROFI},
    {"flp",  VIEW64_FLIGRAPH},
    {"bml",  VIEW64_FLIGRAPH},
    {"flg",  VIEW64_FLIGRAPH},
    {"che",  VIEW64_CHEESE_PAINT},
    {"cwg",  VIEW64_GARFIELD},
    {"dol",  VIEW64_DOLPHIN},
    {"mil",  VIEW64_MICRO_ILLUSTRATOR},
    {"a64",  VIEW64_ARTIST64},
    {"sar",  VIEW64_SARACEN_PAINT},
    {"hed",  VIEW64_HIEDDI},
    {"mui",  VIEW64_MUIFLI},
    {"nup",  VIEW64_NUFLI},
    {"nuf",  VIEW64_NUFLI},
    {"zom",  VIEW64_ZOOMATIC},
    {"fbi",  VIEW64_FLIP},
    {"hie",  VIEW64_HIRES_INTERLACE},
    {"het",  VIEW64_HIRES_EDITOR},
    {"fed",  VIEW64_FLI_EDITOR},
    {"bed",  VIEW64_BITMAP_EDITOR},
    {"fnt",  VIEW64_FONT},
    {"flm",  VIEW64_FLIMATIC},
    {"esh",  VIEW64_EXTEND_SUPER_HIRES_INTERLACE},
    {"sh2",  VIEW64_SUPER_HIRES_II},
    {"sh1",  VIEW64_SUPER_HIRES_I},
    {"emc",  VIEW64_EMC_EDITOR},
    {"hle",  VIEW64_HIRESLACE_EDITOR},
    {"shs",  VIEW64_SUPER_HIRES_STUDIO},
    {"wig",  VIEW64_WIGMORE_ARTIST},
    {"4bt",  VIEW64_GODOT},
    {"clp",  VIEW64_GODOT},
    {"vbm",  VIEW64_VBM},
    {"rpm",  VIEW64_RUN_PAINT_MULTI},
    {"rph",  VIEW64_RUN_PAINT_HIRES},
    {"rpo",  VIEW64_RUN_PAINT_MONO},
    {"p64",  VIEW64_PICASSO64},
    {"fly",  VIEW64_FLYING_COLORS},
    {"lp3",  VIEW64_LOGO_PAINTER_V3},
    {"bfl",  VIEW64_BIGFLI},
    {"bfli", VIEW64_BIGFLI},
    {"cle",  VIEW64_CENTAURI_LOGO_EDITOR},
    {"ile",  VIEW64_INTERLACED_LOGO_EDITOR},
    {"ctm",  VIEW64_CHARPAD},
    {"spd",  VIEW64_SPRITEPAD},
    { NULL,  VIEW64_UNKNOWN}
};

const char *view64_format_name(view64_format format)
{
    if (format >= VIEW64_INVALID || format < VIEW64_UNKNOWN) format = VIEW64_UNKNOWN;
    return view64_format_names[format];
}

view64_format view64_format_detect(const char *name, const unsigned char *mem, size_t size)
{
    int i, l;
    char ext[5] = {0, 0, 0, 0, 0};
    unsigned short int start;

    l = strlen(name);
    if (l > 5 && name[l - 5] == '.') {
        strcpy(ext, name + l - 4);
    }
    if (l > 4 && name[l - 4] == '.') {
        strcpy(ext, name + l - 3);
    }
    if (l > 3 && name[l - 3] == '.') {
        strcpy(ext, name + l - 2);
    }
    for (i = 0; i < 4; i++) if (ext[i] >= 'A' && ext[i] <= 'Z') ext[i] |= 0x20;
    
    i = 0;
    while (view64_extensions[i].ext != NULL) {
        if (!strcmp(view64_extensions[i].ext, ext)) break;
        i++;
    }
    if (view64_extensions[i].format != VIEW64_UNKNOWN) {
        return view64_extensions[i].format;
    }

    if (!strcmp(ext, "she")) {
        if (size == 3250) {
            return VIEW64_SUPER_HIRES;
        } else {
            return VIEW64_SUPER_HIRES_2;
        }
    }

    if (!strcmp(ext, "gig")) {
        if (size == 9002) {
            return VIEW64_GIGA_PAINT_HIRES;
        } else {
            return VIEW64_GIGA_PAINT_MULTI;
        }
    }

    if (!strcmp(ext, "ish")) {
        if (size == 9194) {
            return VIEW64_IMAGE_SYSTEM_HIRES;
        } else {
            return VIEW64_INTERLACE_SUPER_HIRES;
        }
    }
    if (!strcmp(ext, "pmg")) {
        if (size == 9332) {
            return VIEW64_PAINT_MAGIC;
        } else {
            return VIEW64_PAINT_MAGIC_PLUS;
        }
    }

    start = mem[0] + mem[1] * 256;

    if (start == 0x3ff0 && !memcmp(mem + 2, "FUNPAINT (MT) ", 14)) {
        return VIEW64_FUNPAINT;
    } else if (start == 0x5800 && (!memcmp(mem + 2, "DRAZPAINT 1.3", 13) || !memcmp(mem + 2, "DRAZPAINT 1.4", 13) || !memcmp(mem + 2, "DRAZPAINT 2.0", 13))) {
        return VIEW64_DRAZPAINT;
    } else if (start == 0x5800 && !memcmp(mem + 2, "DRAZLACE! 1.0", 13)) {
        return VIEW64_DRAZLACE;
    } else if (start == 0x38f0 && mem[3] == ((0x38ed + size) & 0xff) && mem[4] == ((0x38ed + size) >> 8) && mem[5] == 0x3f && mem[6] == 0x7f) {
        return VIEW64_FLIP;
    } else if (start == 0x3f8e && !memcmp(mem + 6, "\x9e\x32\x30", 3) && size == 9332) {
        return VIEW64_PAINT_MAGIC;
    } else if (start == 0x3f8e && !memcmp(mem + 6, "\x9e\x32\x30", 3) && size == 10332) {
        return VIEW64_PAINT_MAGIC_PLUS;
    } else if (start == 0x4000 && mem[2] == ((0x3ffd + size) & 0xff) && mem[3] == ((0x3ffd + size) >> 8) && mem[4] == 0xf2 && mem[5] == 0x7f) {
        return VIEW64_HIRES_MANAGER;
    } else if (start == 0x0801 && mem[0x75] == 0x4c && mem[0x76] == 0x7e && mem[0x77] == 0x1b && !memcmp(mem + 2, "\x0b\x08\x09\x00\x9e\x32\x30\x35\x39", 9)) {
        return VIEW64_TRUEPAINT;
    } else if (!memcmp(mem, "GOD", 3) && (mem[3] == '0' || mem[3] == '1')) {
        return VIEW64_GODOT;
    } else if (!memcmp(mem, "BM\xcb", 3) && (mem[3] == 2 || mem[3] == 3)) {
        return VIEW64_VBM;
    } else if (start == 0x3aff && mem[2] == 0x66) {
        return VIEW64_FLASH_FLI;
    } else if (start == 0x3bfc && mem[2] == 0x10 && mem[3] == 0x10 && mem[4] == 0x10) {
        return VIEW64_PIXELPERFECT;
    } else if (!memcmp(mem, "CTM", 3) && mem[3] > 0 && mem[3] < 6) {
        return VIEW64_CHARPAD;
    } else if (!memcmp(mem, "SPD", 3) && mem[3] == 1) {
        return VIEW64_SPRITEPAD;
    }

#define E(a, b) ((unsigned)a * 65536 + b)
    switch (size * 65536 + start) {
    case E(3250,  0x9000): return VIEW64_SUPER_HIRES_2;
    case E(4098,  0x1800): return VIEW64_LOGO_PAINTER_V3;
    case E(4098,  0x2000): return VIEW64_INTERLACED_LOGO_EDITOR;
    case E(4174,  0x1800): return VIEW64_LOGO_PAINTER_V3;
    case E(8002,  0x6000): return VIEW64_RUN_PAINT_MONO;
    case E(8194,  0x6000): return VIEW64_CENTAURI_LOGO_EDITOR;
    case E(8642,  0x9000): return VIEW64_SUPER_HIRES;
    case E(9002,  0x4000): return VIEW64_INTERPAINT_HIRES;
    case E(9002,  0x6000): return VIEW64_GIGA_PAINT_HIRES;
    case E(9009,  0x2000): return VIEW64_ART_STUDIO_HIRES;
    case E(9026,  0x5c00): return VIEW64_RUN_PAINT_HIRES;
    case E(9194,  0x4000): return VIEW64_IMAGE_SYSTEM_HIRES;
    case E(9217,  0x5c00): return VIEW64_HIRES_EDITOR;
    case E(9218,  0x2000): return VIEW64_HIEDDI;
    case E(9218,  0x5c00): return VIEW64_DOODLE;
    case E(10003, 0x4000): return VIEW64_INTERPAINT_MULTI;
    case E(10003, 0x6000): return VIEW64_KOALA_PAINTER;// koala painter / paint magic jr. / giga paint multi / run paint multi
    case E(10004, 0x4000): return VIEW64_FACEPAINT;
    case E(10007, 0x8000): return VIEW64_GARFIELD;
    case E(10018, 0x2000): return VIEW64_ART_STUDIO_MULTI;
    case E(10022, 0x18dc): return VIEW64_MICRO_ILLUSTRATOR;
    case E(10050, 0x1800): return VIEW64_PICASSO64; // flying colors
    case E(10050, 0x5800): return VIEW64_VIDCOM64;
    case E(10051, 0x5800): return VIEW64_DRAZPAINT; // cenimate
    case E(10218, 0x3c00): return VIEW64_IMAGE_SYSTEM_MULTI;
    case E(10219, 0x7800): return VIEW64_SARACEN_PAINT;
    case E(10241, 0x5800): return VIEW64_BITMAP_EDITOR;
    case E(10242, 0x0000): return VIEW64_BLAZING_PADDLES;// buggy 64view pictures...
    case E(10242, 0xa000): return VIEW64_BLAZING_PADDLES;
    case E(10242, 0x4000): return VIEW64_WIGMORE_ARTIST;// Artist64
    case E(10242, 0x5800): return VIEW64_DOLPHIN;
    case E(10242, 0x5c00): return VIEW64_RAINBOW_PAINTER;
    case E(10277, 0x7eef): return VIEW64_CDUPAINT;
    case E(10332, 0x3f8e): return VIEW64_PAINT_MAGIC_PLUS;
    case E(12148, 0x5000): return VIEW64_HCB;
    case E(14338, 0x4000): return VIEW64_SUPER_HIRES_STUDIO;
    case E(14770, 0x4650): return (mem[0x1a2] == 0x88) ? VIEW64_SUPER_HIRES_I : VIEW64_SUPER_HIRES_II;
    case E(15355, 0x7fff): return VIEW64_SUPER_HIRES_INTERLACE;
    case E(15362, 0x4000): return VIEW64_SHFXL;
    case E(15874, 0x4000): return VIEW64_SUPER_HIRES_FLI;
    case E(16194, 0x4000): return VIEW64_UFLI;
    case E(16385, 0x4000): return (mem[0x3ff2] == 3) ? VIEW64_HIRES_MANAGER : VIEW64_AFLI;
    case E(16386, 0x4000): return VIEW64_HIRES_FLI_DESIGN;
    case E(17218, 0x3c00): return VIEW64_FLIP;
    case E(17409, 0x3c00): return VIEW64_FLI_DESIGNER;
    case E(17410, 0x0000): return VIEW64_FLI; //buggy 64view pictures...
    case E(17410, 0x3c00): return VIEW64_FLIMATIC;
    case E(17412, 0x4000): return VIEW64_EMC_EDITOR;
    case E(17474, 0x3b00): return VIEW64_FLIGRAPH;
    case E(17665, 0x3b00): return VIEW64_FLI_EDITOR;
    case E(18242, 0x0000): return VIEW64_DRAZLACE; // buggy 64view pictures
    case E(18242, 0x5800): return VIEW64_DRAZLACE;
    case E(18370, 0x3780): return VIEW64_FLIPROFI;
    case E(19434, 0x9c00): return VIEW64_TRUEPAINT;// multicolor interlace editor v1.0, truepaint i
    case E(19922, 0x4000): return VIEW64_UIFLI;
    case E(20454, 0x7fff): return VIEW64_EXTEND_SUPER_HIRES_INTERLACE;
    case E(20482, 0x8000): return VIEW64_CHEESE_PAINT;
    case E(21826, 0x2100): return VIEW64_MUFLI;
    case E(23042, 0x2000): return VIEW64_NUFLI;
    case E(24578, 0x2000): return VIEW64_HIRES_INTERLACE;
    case E(26115, 0x3aff): return VIEW64_FLASH_FLI;
    case E(30738, 0x4000): return VIEW64_INTERLACE_SUPER_HIRES;
    case E(32002, 0x4000): return VIEW64_GODOT;
    case E(32770, 0x4000): return (mem[0x23fa] == 0x7d) ? VIEW64_ECI : VIEW64_HIRESLACE_EDITOR;
    case E(33602, 0x3c00): return VIEW64_PIXELPERFECT;
    case E(33602, 0x0000): return VIEW64_GUNPAINT; //buggy 64view pictures...
    case E(33602, 0x4000): return VIEW64_GUNPAINT; //buggy 64view pictures...
    case E(33603, 0x4000): return VIEW64_GUNPAINT;
    case E(33694, 0x3ff0): return VIEW64_FUNPAINT;
    case E(33795, 0x3bff): return VIEW64_BIGFLI;
    case E(44034, 0x2100): return VIEW64_MUIFLI;
    default: break;
    }
#undef E

    return VIEW64_UNKNOWN;
}

view64_screen *view64_decode(view64_format format, const unsigned char *mem, size_t size, unsigned int flags)
{
    switch (format) {
    case VIEW64_GODOT: return PIC_godot(mem, size);
    case VIEW64_VBM: return PIC_vbm(mem, size);
    case VIEW64_FLI_DESIGNER:
    case VIEW64_FLI: return PIC_fli(mem);
    case VIEW64_EMC_EDITOR: return PIC_emc_editor(mem);
    case VIEW64_FLI_EDITOR: return PIC_fli_editor(mem);
    case VIEW64_FLIGRAPH: return PIC_fligraph(mem, size);
    case VIEW64_FLIP: return PIC_flip(mem, size);
    case VIEW64_FLIPROFI: return PIC_fliprofi(mem);
    case VIEW64_FLASH_FLI: return PIC_flash_fli(mem);
    case VIEW64_HCB: return PIC_hcb(mem);
    case VIEW64_SUPER_HIRES_FLI: return PIC_super_hires_fli(mem, size);
    case VIEW64_XFLI: return PIC_xfli(mem, size);
    case VIEW64_UFLI: return PIC_ufli(mem, size);
    case VIEW64_MUFLI: return PIC_mufli(mem, size);
    case VIEW64_NUFLI: return PIC_nufli(mem, size);
    case VIEW64_MUIFLI: return PIC_muifli(mem, size);
    case VIEW64_UIFLI: return PIC_uifli(mem, size);
    case VIEW64_SHFXL: return PIC_shfxl(mem, size);
    case VIEW64_SUPER_HIRES_INTERLACE_FLI: return PIC_super_hires_interlace_fli(mem, size);
    case VIEW64_SUPER_HIRES_INTERLACE: return PIC_super_hires_interlace(mem, size);
    case VIEW64_EXTEND_SUPER_HIRES_INTERLACE: return PIC_extend_super_hires_interlace(mem, size);
    case VIEW64_SUPER_HIRES_STUDIO: return PIC_super_hires_studio(mem);
    case VIEW64_SUPER_HIRES: return PIC_super_hires(mem);
    case VIEW64_SUPER_HIRES_2: return PIC_super_hires_2(mem);
    case VIEW64_SUPER_HIRES_II: return PIC_super_hires_ii(mem, size);
    case VIEW64_SUPER_HIRES_I: return PIC_super_hires_i(mem, size);
    case VIEW64_FUCKPAINT: return PIC_fuckpaint(mem);
    case VIEW64_TRUEPAINT: return PIC_truepaint(mem, size);
    case VIEW64_GUNPAINT: return PIC_gunpaint(mem);
    case VIEW64_WIGMORE_ARTIST: return PIC_wigmore_artist(mem);
    case VIEW64_ART_STUDIO_MULTI: return PIC_art_studio_multi(mem);
    case VIEW64_ART_STUDIO_MULTI_WINDOW: return PIC_art_studio_multi_window(mem, size);
    case VIEW64_ZOOMATIC: return PIC_zoomatic(mem, size);
    case VIEW64_DRAZPAINT: return PIC_drazpaint(mem, size);
    case VIEW64_DRAZLACE: return PIC_drazlace(mem, size);
    case VIEW64_FUNPAINT: return PIC_funpaint(mem, size);
    case VIEW64_PIXELPERFECT: return PIC_pixelperfect(mem, size);
    case VIEW64_FACEPAINT: return PIC_facepaint(mem);
    case VIEW64_INTERPAINT_HIRES:
    case VIEW64_GIGA_PAINT_HIRES:
    case VIEW64_ART_STUDIO_HIRES: return PIC_art_studio_hires(mem, size);
    case VIEW64_FONT: return PIC_font(mem, size);
    case VIEW64_HIRES_INTERLACE: return PIC_hires_interlace(mem);
    case VIEW64_HIRESLACE_EDITOR: return PIC_hireslace_editor(mem);
    case VIEW64_DOODLE: return PIC_doodle(mem, size);
    case VIEW64_IMAGE_SYSTEM_HIRES: return PIC_image_system_hires(mem);
    case VIEW64_HIEDDI: return PIC_hieddi(mem, size);
    case VIEW64_HIRES_EDITOR: return PIC_hires_editor(mem);
    case VIEW64_IMAGE_SYSTEM_MULTI: return PIC_image_system_multi(mem);
    case VIEW64_VIDCOM64: return PIC_vidcom64(mem);
    case VIEW64_CDUPAINT: return PIC_cdupaint(mem);
    case VIEW64_INTERPAINT_MULTI:
    case VIEW64_GIGA_PAINT_MULTI:
    case VIEW64_KOALA_PAINTER: return PIC_koala_painter(mem, size);
    case VIEW64_BLAZING_PADDLES: return PIC_blazing_paddles(mem);
    case VIEW64_ARTIST64: return PIC_artist64(mem);
    case VIEW64_CHEESE_PAINT: return PIC_cheese_paint(mem);
    case VIEW64_GARFIELD: return PIC_garfield(mem);
    case VIEW64_SARACEN_PAINT: return PIC_saracen_paint(mem);
    case VIEW64_DOLPHIN: return PIC_dolphin(mem);
    case VIEW64_MICRO_ILLUSTRATOR: return PIC_micro_illustrator(mem, size);
    case VIEW64_AFLI: return PIC_afli(mem);
    case VIEW64_FLIMATIC: return PIC_flimatic(mem, size);
    case VIEW64_HIRES_MANAGER: return PIC_hires_manager(mem, size);
    case VIEW64_HIRES_FLI_DESIGN: return PIC_hires_fli_design(mem);
    case VIEW64_ECI: return PIC_eci(mem, size);
    case VIEW64_AMICA: return PIC_amica(mem, size);
    case VIEW64_PAINT_MAGIC: return PIC_paint_magic(mem);
    case VIEW64_BITMAP_EDITOR: return PIC_bitmap_editor(mem);
    case VIEW64_RAINBOW_PAINTER: return PIC_rainbow_painter(mem);
    case VIEW64_RUN_PAINT_MULTI: return PIC_run_paint_multi(mem);
    case VIEW64_RUN_PAINT_HIRES: return PIC_run_paint_hires(mem);
    case VIEW64_RUN_PAINT_MONO: return PIC_run_paint_mono(mem);
    case VIEW64_INTERLACE_SUPER_HIRES: return PIC_interlace_super_hires(mem);
    case VIEW64_FLYING_COLORS:
    case VIEW64_PICASSO64: return PIC_picasso64(mem);
    case VIEW64_LOGO_PAINTER_V3: return PIC_logo_painter(mem, size);
    case VIEW64_BIGFLI: return PIC_bigfli(mem);
    case VIEW64_CENTAURI_LOGO_EDITOR: return PIC_centauri_logo_editor(mem);
    case VIEW64_INTERLACED_LOGO_EDITOR: return PIC_interlaced_logo_editor(mem);
    case VIEW64_PAINT_MAGIC_PLUS: return PIC_paint_magic_plus(mem);
    case VIEW64_CHARPAD: return PIC_charpad(mem, size);
    case VIEW64_SPRITEPAD: return PIC_spritepad(mem, size);
    default: break;
    }
    return NULL;
}
